/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// ============================================================================
#ifndef DETDESC_HISTOPARAM_H
#define DETDESC_HISTOPARAM_H 1
// ============================================================================
// Include files
// ============================================================================
#include "DetDesc/Param.h"
// ============================================================================
// ROOT
// ============================================================================
#ifdef __INTEL_COMPILER // Disable ICC remark from Root headers
#  pragma warning( push )
#  pragma warning( disable : 2259 ) // non-pointer conversion may lose significant bits
#endif
#include "TH1D.h"
#include "TH2D.h"
#include "TH3D.h"
#ifdef __INTEL_COMPILER // Disable ICC remark from Root headers
#  pragma warning( pop )
#endif

namespace DetDesc {
  namespace Params {
    typedef TH1D Histo1D;
    typedef TH2D Histo2D;
    typedef TH3D Histo3D;

    // ========================================================================
    /**  Interface function converting 1d histogram to xml string
     *   @see class Param
     *   @see class TH1D
     *   @param  histo     (INPUT) histogram
     *   @param  name      (INPUT) XML-tag name
     *   @param  comment   (INPUT) the comment
     *   @param  precision (INPUT) not used
     *   @return XML-string
     *   @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl, Dmitry Golubkov
     *   @date 2009-12-06
     */
    GAUDI_API std::string toXMLStr( const DetDesc::Params::Histo1D& histo, std::string_view name,
                                    std::string_view comment, int precision );

    // ========================================================================
    /**  Interface function converting 2d histogram to xml string
     *   @see class Param
     *   @see class TH2D
     *   @param  histo     (INPUT) histogram
     *   @param  name      (INPUT) XML-tag name
     *   @param  comment   (INPUT) the comment
     *   @param  precision (INPUT) not used
     *   @return XML-string
     *   @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl, Dmitry Golubkov
     *   @date 2009-12-06
     */
    GAUDI_API std::string toXMLStr( const DetDesc::Params::Histo2D& histo, std::string_view name,
                                    std::string_view comment, int precision );

    // ========================================================================
    /**  Interface function converting 3d histogram to xml string
     *   @see class Param
     *   @see class TH3D
     *   @param  histo     (INPUT) histogram
     *   @param  name      (INPUT) XML-tag name
     *   @param  comment   (INPUT) the comment
     *   @param  precision (INPUT) not used
     *   @return XML-string
     */
    GAUDI_API std::string toXMLStr( const DetDesc::Params::Histo3D& histo, std::string_view name,
                                    std::string_view comment, int precision );

    // ========================================================================
    /**  Interface function converting 1d histogram to a custom format string
     *   @see class Param
     *   @see class TH1D
     *   @param  histo     (INPUT) histogram
     *   @return string
     *   @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl, Dmitry Golubkov
     *   @date 2009-12-06
     */
    GAUDI_API std::string toStr( const DetDesc::Params::Histo1D& histo );

    // ========================================================================
    /**  Interface function converting 2d histogram to a custom format string
     *   @see class Param
     *   @see class TH2D
     *   @param  histo     (INPUT) histogram
     *   @return string
     *   @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl, Dmitry Golubkov
     *   @date 2009-12-06
     */
    GAUDI_API std::string toStr( const DetDesc::Params::Histo2D& histo );

    // ========================================================================
    /**  Interface function converting 3d histogram to a custom format string
     *   @see class Param
     *   @see class TH3D
     *   @param  histo     (INPUT) histogram
     *   @return string
     */
    GAUDI_API std::string toStr( const DetDesc::Params::Histo3D& histo );

  } // namespace Params
} // namespace DetDesc
// ============================================================================
/**  The template specialization of Param methods for the histograms
 *   @see class Param
 *   @see class TH1D
 *   @param  name    (INPUT) XML-tag name
 *   @param  comment (INPUT the comment
 *   @param  precision (INPUT) not used
 *   @return XML-string
 *   @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl
 *   @date 2009-10-22
 */
template <>
inline std::string Param<DetDesc::Params::Histo1D>::toXMLStr( std::string_view name, std::string_view comment,
                                                              int precision ) const {
  return DetDesc::Params::toXMLStr( m_val, name, comment, precision );
}
// ============================================================================
/**  The template specialization of Param methods for the histograms
 *   @see class Param
 *   @see class TH1D
 *   @return string representation of the histogram
 *   @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl
 *   @date 2009-10-22
 */
template <>
inline std::string Param<DetDesc::Params::Histo1D>::toStr() const {
  return DetDesc::Params::toStr( m_val );
}
// ============================================================================
/**  The template specialization of Param methods for the histograms
 *   @see class Param
 *   @see class TH2D
 *   @param  name     (INPUT) XML-tag name
 *   @param  comment  (INPUT the comment
 *   @param  precision (INPUT) not used
 *   @return XML-string
 *   @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl
 *   @date 2009-10-22
 */
template <>
inline std::string Param<DetDesc::Params::Histo2D>::toXMLStr( std::string_view name, std::string_view comment,
                                                              int precision ) const {
  return DetDesc::Params::toXMLStr( m_val, name, comment, precision );
}
// ============================================================================
/**  The template specialization of Param methods for the histograms
 *   @see class Param
 *   @see class TH2D
 *   @return string representation of the histogram
 *   @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl
 *   @date 2009-10-22
 */
template <>
inline std::string Param<DetDesc::Params::Histo2D>::toStr() const {
  return DetDesc::Params::toStr( m_val );
}
// ============================================================================
/**  The template specialization of Param methods for the histograms
 *   @see class Param
 *   @see class TH3D
 *   @param  name     (INPUT) XML-tag name
 *   @param  comment  (INPUT the comment
 *   @param  precision (INPUT) not used
 *   @return XML-string
 */
template <>
inline std::string Param<DetDesc::Params::Histo3D>::toXMLStr( std::string_view name, std::string_view comment,
                                                              int precision ) const {
  return DetDesc::Params::toXMLStr( m_val, name, comment, precision );
}
// ============================================================================
/**  The template specialization of Param methods for the histograms
 *   @see class Param
 *   @see class TH3D
 *   @return string representation of the histogram
 */
template <>
inline std::string Param<DetDesc::Params::Histo3D>::toStr() const {
  return DetDesc::Params::toStr( m_val );
}
// ============================================================================
// The END
// ============================================================================
#endif // DETDESC_HISTOPARAM_H
