/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

// Include files

#include "DetDesc/IDetectorElement.h"
#include "DetDesc/IGeometryInfo.h"
// local
#include "DetDesc/3DTransformationFunctions.h"
#include "DetDesc/GlobalToLocalDelta.h"
namespace DetDesc {

  const Gaudi::Transform3D localDeltaMatrix( const IDetectorElementPlus* DE, const Gaudi::Transform3D& globalDelta ) {
    return localDeltaMatrix( DE->geometryPlus(), globalDelta );
  }

  const Gaudi::Transform3D localDeltaMatrix( const IGeometryInfoPlus*  GI,
                                             const Gaudi::Transform3D& toMisaligned_global ) {

    const Gaudi::Transform3D d_0 = GI->ownToOffNominalMatrix();
    const Gaudi::Transform3D T   = GI->toGlobalMatrix();
    return d_0 * T.Inverse() * toMisaligned_global * T;
  }

  void localDeltaParameters( const IDetectorElementPlus* DE, std::vector<double>& trans, std::vector<double>& rot,
                             std::vector<double>& pivot ) {

    DetDesc::localDeltaParameters( DE->geometryPlus(), trans, rot, pivot );
  }

  void localDeltaParameters( const IGeometryInfoPlus* GI, std::vector<double>& trans, std::vector<double>& rot,
                             std::vector<double>& pivot ) {

    const Gaudi::Transform3D globalDelta = DetDesc::localToGlobalTransformation( trans, rot, pivot );
    const Gaudi::Transform3D localDelta  = DetDesc::localDeltaMatrix( GI, globalDelta );
    // get the pivot from the alignment condition
    // @todo find a way not to require this awful hard-wired string.
    const std::string pivotString( "pivotXYZ" );

    std::vector<double> tmp = ( GI->alignmentCondition()->exists( pivotString ) )
                                  ? GI->alignmentCondition()->paramAsDoubleVect( pivotString )
                                  : std::vector<double>( 3, 0 );

    std::copy( tmp.begin(), tmp.end(), pivot.begin() );

    DetDesc::getZYXTransformParameters( localDelta, trans, rot, pivot );
  }

  const Gaudi::Transform3D localDeltaFromGlobalTransform( const IDetectorElementPlus* DE,
                                                          const Gaudi::Transform3D&   globalTransform ) {
    return localDeltaFromGlobalTransform( DE->geometryPlus(), globalTransform );
  }

  const Gaudi::Transform3D localDeltaFromGlobalTransform( const IGeometryInfoPlus*  GI,
                                                          const Gaudi::Transform3D& globalTransform ) {
    const Gaudi::Transform3D d_0   = GI->ownToOffNominalMatrix();
    const Gaudi::Transform3D T_inv = GI->toLocalMatrix();
    return d_0 * T_inv * globalTransform;
  }

  const Gaudi::Transform3D localDeltaFromParentTransform( const IDetectorElementPlus* parentDE,
                                                          const IDetectorElementPlus* DE,
                                                          const Gaudi::Transform3D&   parentTransform ) {
    return localDeltaFromParentTransform( parentDE->geometryPlus(), DE->geometryPlus(), parentTransform );
  }

  const Gaudi::Transform3D localDeltaFromParentTransform( const IGeometryInfoPlus*  parentGI,
                                                          const IGeometryInfoPlus*  GI,
                                                          const Gaudi::Transform3D& parentTransform ) {
    const Gaudi::Transform3D d_0       = GI->ownToOffNominalMatrix();
    const Gaudi::Transform3D T_parent  = parentGI->toGlobalMatrix();
    const Gaudi::Transform3D T_k2n_inv = GI->toLocalMatrix() * T_parent;
    return d_0 * T_k2n_inv * parentTransform;
  }

  const Gaudi::Transform3D localDeltaFromAnyFrame( const Gaudi::Transform3D& frame, const IDetectorElementPlus* DE,
                                                   const Gaudi::Transform3D& toGlobalInFrame ) {
    return localDeltaFromAnyFrame( frame, DE->geometryPlus(), toGlobalInFrame );
  }

  const Gaudi::Transform3D localDeltaFromAnyFrame( const Gaudi::Transform3D& frame, const IGeometryInfoPlus* GI,
                                                   const Gaudi::Transform3D& toGlobalInFrame ) {
    const Gaudi::Transform3D d_0       = GI->ownToOffNominalMatrix();
    const Gaudi::Transform3D T_k2n_inv = GI->toLocalMatrix() * frame;
    return d_0 * T_k2n_inv * toGlobalInFrame;
  }

  const Gaudi::Transform3D localDeltaFromAnyFrameDelta( const Gaudi::Transform3D& frame, const IDetectorElementPlus* DE,
                                                        const Gaudi::Transform3D& deltaInFrame ) {
    return localDeltaFromAnyFrameDelta( frame, DE->geometryPlus(), deltaInFrame );
  }

  const Gaudi::Transform3D localDeltaFromAnyFrameDelta( const Gaudi::Transform3D& frame, const IGeometryInfoPlus* GI,
                                                        const Gaudi::Transform3D& deltaInFrame ) {
    const Gaudi::Transform3D d_0 = GI->ownToOffNominalMatrix();
    return d_0 * frame.Inverse() * deltaInFrame * frame;
  }

  const Gaudi::Transform3D globalDeltaFromLocalDeltaPrime( const IDetectorElementPlus* DE,
                                                           const Gaudi::Transform3D&   deltaPrime ) {
    return globalDeltaFromLocalDeltaPrime( DE->geometryPlus(), deltaPrime );
  }

  const Gaudi::Transform3D globalDeltaFromLocalDeltaPrime( const IGeometryInfoPlus*  GI,
                                                           const Gaudi::Transform3D& deltaPrime ) {
    return GI->toGlobalMatrix() * deltaPrime * GI->toLocalMatrix();
  }

  const Gaudi::Transform3D globalDelta( const IDetectorElementPlus* DE ) { return globalDelta( DE->geometryPlus() ); }

  const Gaudi::Transform3D globalDelta( const IGeometryInfoPlus* GI ) {
    return GI->toGlobalMatrix() * GI->ownToOffNominalMatrix() * GI->toLocalMatrix();
  }

  const Gaudi::Transform3D relativeTransformation( const IDetectorElementPlus* referenceDE,
                                                   const IDetectorElementPlus* DE ) {
    return referenceDE->geometryPlus()->toLocalMatrix() * DE->geometryPlus()->toGlobalMatrix();
  }

} // namespace DetDesc
