/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#include "MuonDet/MuonChamberLayout.h"
#include "MuonDet/DeMuonDetector.h"
#include "MuonDet/MuonChamberGrid.h"

#include "Core/FloatComparison.h"
#include "DetDesc/DetectorElement.h"
#include "DetDesc/IGeometryInfo.h"
#include "DetDesc/IPVolume.h"
#include "DetDesc/SolidBox.h"

#include "fmt/format.h"

#include <exception>
#include <iostream>
#include <vector>

//-----------------------------------------------------------------------------
// Implementation file for class : MuonChamberLayout
//
// 2004-01-07 : Alessio Sarti
//-----------------------------------------------------------------------------
namespace {
  const Gaudi::XYZPoint origin{ 0, 0, 0 };

  template <typename T>
  auto logic_error( T&& x ) {
    return GaudiException( std::forward<T>( x ), "MuonChamberLayout", StatusCode::FAILURE );
  }
} // namespace

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
// When initialized in this way needs un update for the Layout
MuonChamberLayout::MuonChamberLayout()
    : MuonChamberLayout{ MuonLayout{ 1, 1 }, MuonLayout{ 1, 2 }, MuonLayout{ 1, 4 }, MuonLayout{ 2, 8 }, false } {

  throw GaudiException( "problem", "MuonChamberLayout::MuonChamberLayout", StatusCode::FAILURE );
}

MuonChamberLayout::MuonChamberLayout( MuonLayout R1, MuonLayout R2, MuonLayout R3, MuonLayout R4, bool isM1defined ) {
  m_isM1defined = isM1defined;
  // Setting up the layout for the various regions
  m_layout[0] = R1;
  m_layout[1] = R2;
  m_layout[2] = R3;
  m_layout[3] = R4;
  // Performing vectors and grid initialization
  initialize().ignore( /* AUTOMATICALLY ADDED FOR gaudi/Gaudi!763 */ );
}

StatusCode MuonChamberLayout::initialize() {
  return DetectorElementPlus::initialize().andThen( [&] {
    // GPGP
    // Grid initialization N.B. these numbers are already defined in the MuonLayout
    // why hard-code them here ????
    int          iDd;
    unsigned int cgX[4] = { 1, 1, 1, 2 };
    m_cgX.resize( 4 );
    for ( iDd = 0; iDd < 4; iDd++ ) { m_cgX.at( iDd ) = cgX[iDd]; }
    unsigned int cgY[4] = { 1, 2, 4, 8 };
    m_cgY.resize( 4 );
    for ( iDd = 0; iDd < 4; iDd++ ) { m_cgY.at( iDd ) = cgY[iDd]; }
    // GPGP also these constants can be obtained from the #of chambers/region
    // and the layout
    // Offset: needs to change?
    int myoffset[4] = { 0, 16, 48, 112 };
    m_offSet.resize( 4 );
    for ( iDd = 0; iDd < 4; iDd++ ) { m_offSet.at( iDd ) = myoffset[iDd]; }

    // Initializing the chamber grid with detector elements
    m_chamberGrid.resize( 368 );
    for ( int iD = 0; iD < 368; iD++ ) { m_chamberGrid.at( iD ) = -1; }
    return StatusCode::SUCCESS;
  } );
}

void MuonChamberLayout::Copy( MuonChamberLayout& lay ) {
  // Copy constructor
  for ( int iL = 0; iL < 4; iL++ ) { lay.setLayout( iL, this->layout( iL ) ); }
}

std::vector<int> MuonChamberLayout::neighborChambers( float myX, float myY, int stat, int x_direction,
                                                      int y_direction ) const {
  std::vector<int> myChambers;
  int              sC_idY( 0 ), sC_idX( 0 ), reg( 0 );
  gridPosition( myX, myY, stat, sC_idX, sC_idY, reg );
  // Deciding the corner-side
  int scX( 0 ), scY( 0 );
  if ( x_direction != 0 ) scX++;
  if ( y_direction != 0 ) scY++;

  for ( int ix = 0; ix <= scX; ix++ ) {
    for ( int iy = 0; iy <= scY; iy++ ) {
      // Do the computation only for OTHER chambers
      if ( ix || iy ) { chamberXY( sC_idX, sC_idY, ix * x_direction, iy * y_direction, reg, myChambers ); }
    }
  }
  return myChambers;
}

void MuonChamberLayout::returnChambers( int sta, float st_x, float st_y, int x_dir, int y_dir, std::vector<int>& regs,
                                        std::vector<int>& chs ) const {
  // Now get the chambers
  for ( const auto& idTile : neighborChambers( st_x, st_y, sta, x_dir, y_dir ) ) {
    const long int cTile = idTile;
    // Added protection against non existing chamber
    if ( cTile > 0 ) {
      int region = findRegion( cTile );
      regs.push_back( region );
      chs.push_back( cTile - m_offSet.at( region ) - 1 );
    }
  }
}

void MuonChamberLayout::chamberXY( int sx, int sy, int shx, int shy, int reg, std::vector<int>& chamberNumber ) const {
  int vSize( 0 );
  int chN = -1;
  int fx  = sx + shx;
  int fy  = sy + shy;

  // Protect against chambers outside the current region grid
  if ( fx < 0 || fy < 0 || fx >= 4 * (int)m_cgX.at( reg ) || fy >= 4 * (int)m_cgY.at( reg ) ) {

    // Starting point in the new (larger) grid
    sy = sy + 2 * m_cgY.at( reg );
    if ( reg < 2 ) {
      if ( sx < 2 * (int)m_cgX.at( reg ) ) {
        sx = 1;
      } else {
        sx = 2;
      }
    }

    // Goes to other region
    if ( reg + 1 < 4 ) {
      sx = ( sx / ( 2 / m_cgX.at( reg + 1 ) ) ) + m_cgX.at( reg + 1 );
      // Checks in larger grid (higher region)
      chamberXY( sx, sy, shx, shy, reg + 1, chamberNumber );
    } else {
      // Exits [chamber outside the R4 grid]
      chamberNumber.push_back( -1 );
      return;
    }
    // Fix the chamber number for chambers in region != from starting one
    vSize = chamberNumber.size();
    if ( vSize > 0 ) {
      if ( chamberNumber.at( vSize - 1 ) > -1 ) return;
    }
  } else {
    // Look inside the grid
    int enc = fx + 4 * m_cgX.at( reg ) * fy + m_offSet.at( reg );
    chN     = m_chamberGrid.at( enc );
  }

  // If a chamber number is <0 this means that
  //-> The chamber is in a different region
  //-> The chamber has NOT been turned on in the xml
  if ( chN < 0 && shouldLowReg( fx, fy, reg ) ) {
    // Found a -1 chamber: need to low
    // the region number (if > 1) to look in inside inner region
    if ( reg - 1 >= 0 ) {
      // When dropping down the region number
      if ( reg < 3 ) {
        if ( fx < 2 * (int)m_cgX.at( reg ) ) {
          shx = -1;
        } else {
          shx = 1;
        }
      } else {
        shx = -2;
      }
      shy = m_cgY.at( reg );
      chamberXY( fx, fy, shx, -shy, reg - 1, chamberNumber );
    } else {
      // lowest (inner) region. Exiting
      chamberNumber.push_back( -1 );
      return;
    }
    vSize = chamberNumber.size();
    if ( vSize > 0 ) {
      if ( chamberNumber.at( vSize - 1 ) > -1 ) { chN = chamberNumber.at( vSize - 1 ); }
    }

    // Second chamber needed for regions 0,1,2
    if ( reg < 3 ) {
      if ( ( sy / m_cgY.at( reg ) < 1 ) || ( sy / m_cgY.at( reg ) > 2 ) ) {
        if ( reg - 1 > 0 ) {
          chamberXY( fx, fy, 0, 0 - m_cgY.at( reg ), reg - 1, chamberNumber );
        } else {
          chamberNumber.push_back( -1 );
          return;
        }
        vSize = chamberNumber.size();
        if ( vSize ) {
          if ( chamberNumber.at( vSize - 1 ) > -1 ) return;
        }
      }
    }
  } else if ( chN != -1 ) {
    // Need a +1 to avoid skipping chamber n.1 (idx 0)
    chamberNumber.push_back( chN + m_offSet.at( reg ) + 1 );
  }
}

int MuonChamberLayout::findRegion( int chamber ) const {
  int reg( 0 );
  int offset[4] = { 16, 48, 112, 368 };
  for ( reg = 0; reg < 4; reg++ ) {
    if ( chamber < offset[reg] ) break;
  }
  if ( reg > 3 ) { reg = -1; }
  return reg;
}

void MuonChamberLayout::chamberMostLikely( float x, float y, int station, int& chmb, int& reg ) const {

  int myReg( -1 );
  int fx( -1 ), fy( -1 );
  if ( !LHCb::essentiallyZero( m_xS.at( station ) ) && !LHCb::essentiallyZero( m_yS.at( station ) ) ) {
    gridPosition( x, y, station, fx, fy, myReg );
  } else {
    throw logic_error( "Null chamber dimensions" );
  }
  if ( myReg < 0 ) {
    reg  = -1;
    chmb = -1;
    return;
  }
  // Here's the chamber!!!!
  if ( fx > -1 && fy > -1 ) {
    // Look inside the grid
    int enc = fx + 4 * m_cgX.at( myReg ) * fy + m_offSet.at( myReg );
    chmb    = m_chamberGrid.at( enc );
    reg     = myReg;
  } else {
    throw logic_error( "Matrix Index problem!!!!" );
  }
}

// Returns the Tile for a given chamber
LHCb::Detector::Muon::TileID MuonChamberLayout::tileChamber( DeMuonChamber* chmb ) const {
  LHCb::Detector::Muon::TileID myTile;

  // Region and station got from chamber
  myTile.setStation( chmb->stationNumber() );
  unsigned int reg = chmb->regionNumber();
  myTile.setRegion( reg );

  // Gap is not used here
  // myTile.setLayer(0);

  myTile.setLayout( MuonLayout( m_cgX.at( reg ), m_cgY.at( reg ) ) );

  // Chamber number is needed to find x,y
  unsigned int chN = chmb->chamberNumber();
  unsigned int mychN;
  unsigned int fx( 0 ), fy( 0 ), enc( 0 );
  for ( fx = 0; fx < 4 * m_cgX.at( reg ); fx++ ) {
    for ( fy = 0; fy < 4 * m_cgY.at( reg ); fy++ ) {
      enc   = fx + 4 * m_cgX.at( reg ) * fy + m_offSet.at( reg );
      mychN = m_chamberGrid.at( enc );
      if ( chN == mychN ) {
        // quarter definition according to D.Hutchcroft
        //      +y
        //    3 | 0
        //   ---+--- +x
        //    2 | 1

        int myQuarter = -1;
        if ( fx >= m_cgX.at( reg ) * 2 ) {
          myQuarter = 1;
          if ( fy >= m_cgY.at( reg ) * 2 ) { myQuarter = 0; }
        } else {
          myQuarter = 2;
          if ( fy >= m_cgY.at( reg ) * 2 ) { myQuarter = 3; }
        }

        if ( myQuarter == 0 ) {
          myTile.setX( fx - m_cgX.at( reg ) * 2 );
          myTile.setY( fy - m_cgY.at( reg ) * 2 );
        } else if ( myQuarter == 1 ) {
          myTile.setX( fx - m_cgX.at( reg ) * 2 );
          myTile.setY( -1 - fy + m_cgY.at( reg ) * 2 );
        } else if ( myQuarter == 2 ) {
          myTile.setX( -1 - fx + m_cgX.at( reg ) * 2 );
          myTile.setY( -1 - fy + m_cgY.at( reg ) * 2 );
        } else if ( myQuarter == 3 ) {
          myTile.setX( -1 - fx + m_cgX.at( reg ) * 2 );
          myTile.setY( fy - m_cgY.at( reg ) * 2 );
        }
        myTile.setQuarter( myQuarter );

        return myTile;
      }
    }
  }
  throw logic_error( "X, Y and Q not found! Check for grid problems" );
}

// Returns the Tile for a given chamber number
LHCb::Detector::Muon::TileID MuonChamberLayout::tileChamberNumber( int sta, int reg, int chmbNum ) const {

  int encode = 276 * sta + chmbNum;
  if ( reg == 1 ) encode = encode + 12;
  if ( reg == 2 ) encode = encode + 12 + 24;
  if ( reg == 3 ) encode = encode + 12 + 24 + 48;
  LHCb::Detector::Muon::TileID myTile;

  DeMuonChamber* deChmb = m_ChVec[encode];
  if ( deChmb ) {
    myTile = tileChamber( deChmb );
  } else {
    throw logic_error( fmt::format( "Could not find chamber {} in the TES. Check the configuration!", chmbNum ) );
  }
  return myTile;
}

void MuonChamberLayout::gridPosition( float x, float y, int iS, int& idx, int& idy, int& reg ) const {

  int   myIdx( 0 ), myIdy( 0 ), myReg( -1 );
  float rX( 0 ), rY( 0 );
  float xs = m_xS.at( iS );
  float ys = m_yS.at( iS );
  // Extracting chamber indexes in the finest grid.
  myIdx = (int)( x / xs );
  myIdy = (int)( y / ys );
  rX    = x - xs * myIdx;
  rY    = y - ys * myIdy;

  for ( int iRx = 0; iRx < 4; iRx++ ) {
    // To be checked fator 2 in last gridX
    if ( abs( myIdx ) < 2 * (int)m_cgX.at( iRx ) && abs( myIdy ) < 2 * (int)m_cgY.at( iRx ) ) {
      myReg = iRx;
      break;
    } else {
      if ( iRx < 2 ) myIdx /= 2;
    }
  }
  if ( myIdy <= 0 && rY < 0 ) myIdy--;
  if ( myIdx <= 0 && rX < 0 ) myIdx--;
  if ( myReg > -1 ) {
    idx = myIdx + 2 * (int)m_cgX.at( myReg );
    idy = myIdy + 2 * (int)m_cgY.at( myReg );
    reg = myReg;
  }
}

// GPGP also these numbers can go to DDDB
// they are station specific constants: chamber sensitive dimensions
void MuonChamberLayout::setGridStep() {
  float dimX[5] = { 240.f, 301.363f, 325.155f, 348.947f, 372.739f };
  float dimY[5] = { 200.f, 251.136f, 270.962f, 290.789f, 310.615f };
  for ( int iDum = ( m_isM1defined ? 0 : 1 ); iDum < 5; iDum++ ) {
    m_xS.push_back( dimX[iDum] );
    m_yS.push_back( dimY[iDum] );
  }
}

std::vector<DeMuonChamber*> MuonChamberLayout::fillChambersVector( DeMuonDetector const& muonDet ) {
  int idx( -1 ), idy( -1 ), reg( -1 );

  // GPGP another hard-coded set: these will be in DDDB
  m_ChVec.resize( 1380 ); // idem

  StatusCode sc = StatusCode::SUCCESS;

  // Getting stations

  // Reset the cache information
  m_logHorizGridX.resize( 20 );
  m_logHorizGridY.resize( 20 );
  m_logVertGridX.resize( 20 );
  m_logVertGridY.resize( 20 );
  m_padGridX.resize( 20 );
  m_padGridY.resize( 20 );
  for ( int ir = 0; ir < 20; ir++ ) {
    m_logHorizGridX.at( ir ) = m_logHorizGridY.at( ir ) = m_logVertGridX.at( ir ) = m_logVertGridY.at( ir ) =
        m_padGridX.at( ir ) = m_padGridY.at( ir ) = 0;
  }

  // Set the grid steps for the various stations
  setGridStep();
  int vIdx( -1 ), myvIdx( -1 );
  int iS( 0 ), obtIS( 0 );
  // int encode( 0 );
  for ( auto* itSt : muonDet.childIDetectorElements() ) {

    // Are there any void Stations?
    std::string name = itSt->name();

    if ( name.find( "/MF" ) != name.npos ) // if it finds a filter, it skips to
      continue;                            // the next element

    int         len   = name.size();
    int         start = DeMuonLocation::Default.size();
    std::string substring;
    substring.assign( name, start, len );

    sscanf( substring.c_str(), "/M%d", &obtIS );

    // while ( iS != obtIS - ( m_isM1defined ? 1 : 2 ) ) {
    //   msgStream() << MSG::WARNING << "There is/are void stations! " << endmsg;
    //   for ( int ire = 0; ire < 4; ire++ ) {
    //     for ( int ich = 0; ich < MaxRegions[ire]; ich++ ) {
    //       m_ChVec.at( encode ) = (DeMuonChamber*)0;
    //       encode++;
    //     }
    //   }
    //   iS++;
    // }

    if ( iS != obtIS - ( m_isM1defined ? 1 : 2 ) ) {
      std::cerr << "iS=" << iS << " obtIS" << obtIS << std::endl;
      throw logic_error( "There is/are void stations!" );
    }
    // getting station side (A or C)
    for ( auto* itSide : itSt->childIDetectorElements() ) {

      // Getting regions
      int iR( 0 ), obtIR( 0 );
      for ( auto* itRg : itSide->childIDetectorElements() ) {

        // Are there any void Regions?
        name  = ( itRg->name() ).c_str();
        start = ( DeMuonLocation::Default ).size();
        std::string substringSta;
        substringSta.assign( name, start, 3 );
        std::string substringReg;
        substringReg.assign( name, start + 11, 3 );

        sscanf( substringSta.c_str(), "/M%d", &obtIS );

        sscanf( substringReg.c_str(), "/R%d", &obtIR );

        if ( iR != obtIR - 1 ) throw logic_error( "There is/are void regions. " );

        // Index needed to fill the cache information
        // There are 4 regions per station: this is not going to change!
        vIdx = iR + iS * 4;
        reg  = iR;
        // Getting chambers
        for ( auto* itCh : itRg->childIDetectorElements() ) {
          DeMuonChamber* deChmb  = dynamic_cast<DeMuonChamber*>( itCh );
          int            countCh = 276 * iS + deChmb->chamberNumber();
          if ( iR == 1 ) countCh = countCh + 12;
          if ( iR == 2 ) countCh = countCh + 12 + 24;
          if ( iR == 3 ) countCh = countCh + 12 + 24 + 48;

          //   if(deChmb->chamberNumber()==4)
          //     std::cout<<" ch position "<<countCh<<" "<<iS<<" "<<iR<<" "<<deChmb->chamberNumber()<<endmsg;
          //    if(deChmb) {
          //      m_ChVec.at(encode) = deChmb;
          //    } else {
          //      m_ChVec.at(encode) = (DeMuonChamber*)0;
          //    }
          m_ChVec.at( countCh ) = deChmb;
          // encode++;

          /* retrieve the chamber grid parameters according to their central XY point
             the chamber position is encoded in each region on a grid of ncham x ncham =
             R1: 4x4  = 16 (16  = m_offset[1])
             R1: 8x4  = 32 (48  = 16+32 = m_offset[2])
             R1: 16x4 = 64 (112 = 48+48 + m_offset[3])
             R1: 32x8  = 256 -> tot = 368
             the origin of the grid is at the lower-right corner (side C, Q3, x,y<0)
             e.g. R1:
                       y

                       A
               SIDE A  |  SIDE C
             +----+----+----+----+
             |ch01|ch02|ch03|ch04|
             | 15 | 14 | 13 | 12 |
             +----+----+----+----+
             |ch05| xx | xx |ch06|
             | 11 | 10 | 09 | 08 |
       x <---+----+----o----+----+
             |ch07| xx | xx |ch08|
             | 07 | 06 | 05 | 04 |
             +----+----+----+----+
             |ch09|ch10|ch11|ch12|
             | 03 | 02 | 01 | 00 |
             +----+----+----+----+ (0,0)
          */

          auto   glob_orig = deChmb->geometryPlus()->toGlobal( origin );
          double myX       = glob_orig.x();
          double myY       = glob_orig.y();
          gridPosition( (float)myX, (float)myY, iS, idx, idy, reg );

          int enc                 = idx + 4 * m_cgX.at( reg ) * idy + m_offSet.at( reg );
          m_chamberGrid.at( enc ) = deChmb->chamberNumber();

          // Try to fill also the other relevant quantities
          // Only when the region changes
          if ( vIdx != myvIdx ) {
            myvIdx = vIdx;
            sc     = fillSystemGrids( deChmb, vIdx, reg );
            if ( sc.isFailure() ) {
              throw logic_error( fmt::format( "Failed to fill the system grid for chamber {} in region {} {}",
                                              deChmb->chamberNumber(), reg, iS ) );
            }
          }
        }
        // next region
        iR++;
      }
      // next side
    }

    // next station
    iS++;
  }
  return m_ChVec;
}

bool MuonChamberLayout::shouldLowReg( int idX, int idY, int reg ) const {
  bool lower = true;
  bool iXnz  = ( ( ( idX < (int)m_cgX.at( reg ) ) && ( idX > 0 ) ) ||
                ( ( idX >= 3 * (int)m_cgX.at( reg ) ) && ( idX < 4 * (int)m_cgX.at( reg ) ) ) );
  bool iYnz  = ( ( ( idY < (int)m_cgY.at( reg ) ) && ( idY > 0 ) ) ||
                ( ( idY >= 4 * (int)m_cgY.at( reg ) ) && ( idX < 4 * (int)m_cgY.at( reg ) ) ) );
  if ( iXnz && iYnz ) { lower = false; }
  return lower;
}

std::optional<DeMuonDetector::TilePosition> MuonChamberLayout::position( const LHCb::Detector::Muon::TileID& tile ) {
  // OK how to make this work.....
  // first locate the station and region from the tile
  unsigned int station = tile.station();
  unsigned int region  = tile.region();

  // now compare the layout parameter to possible "levels"
  // currently chamber, logical channel, pad
  if ( m_cgX.at( region ) == tile.layout().xGrid() && m_cgY.at( region ) == tile.layout().yGrid() ) {
    // chambers
    return getXYZChamberTile( tile, true );
  } else if ( m_padGridX[station * 4 + region] == tile.layout().xGrid() &&
              m_padGridY[station * 4 + region] == tile.layout().yGrid() ) {
    // muon pads and logical channels with a 1:1 mapping to pads
    return getXYZPad( tile );
  } else if ( m_logHorizGridX[station * 4 + region] == tile.layout().xGrid() &&
              m_logHorizGridY[station * 4 + region] == tile.layout().yGrid() ) {
    // horizontal logical channels
    return getXYZLogical( tile );
  } else if ( m_logVertGridX[station * 4 + region] == tile.layout().xGrid() &&
              m_logVertGridY[station * 4 + region] == tile.layout().yGrid() ) {
    // vertical logical channels
    return getXYZLogical( tile );
  } else if ( 1 == tile.layout().xGrid() && 1 == tile.layout().yGrid() ) {
    return getXYZTwelfth( tile );
  }
  throw logic_error( fmt::format( "Did not understand the LHCb::Detector::Muon::TileID encoding xGrid={} yGrid={}, "
                                  "please check the muon geometry version.",
                                  tile.layout().xGrid(), tile.layout().yGrid() ) );
}

StatusCode MuonChamberLayout::fillSystemGrids( DeMuonChamber* deChmb, int vIdx, int reg ) {

  // Getting the grid pointer
  Condition*       aGrid   = deChmb->condition( deChmb->getGridName() );
  MuonChamberGrid* theGrid = dynamic_cast<MuonChamberGrid*>( aGrid );

  if ( !theGrid ) { return StatusCode::FAILURE; }

  std::vector<int> readoutType = theGrid->getReadoutGrid();
  std::vector<int> mapType     = theGrid->getMapGrid();

  // Phys Channels Vectors
  int grX  = m_cgX.at( reg ) * theGrid->getGrid1SizeX();
  int grY  = m_cgY.at( reg ) * theGrid->getGrid1SizeY();
  int SgrX = m_cgX.at( reg ) * theGrid->getGrid2SizeX();
  int SgrY = m_cgY.at( reg ) * theGrid->getGrid2SizeY();

  // Log channels multipl
  int xm0( 0 ), xm1( 0 ), ym0( 0 ), ym1( 0 );
  xm0 = mapType.at( 0 );
  ym0 = mapType.at( 1 );
  if ( mapType.size() == 4 ) {
    xm1 = mapType.at( 2 );
    ym1 = mapType.at( 3 );
  }

  // Service variables
  int axm0( 0 ), axm1( 0 ), aym0( 0 ), aym1( 0 );
  int agrX( 0 ), agrY( 0 ), aSgrX( 0 ), aSgrY( 0 );

  // Only one readout
  if ( readoutType.size() == 1 ) {
    // and only one map
    if ( mapType.size() == 2 ) {
      // so horiz, vert and pad all same
      m_logHorizGridX.at( vIdx ) = grX / xm0;
      m_logHorizGridY.at( vIdx ) = grY / ym0;
      m_logVertGridX.at( vIdx )  = grX / xm0;
      m_logVertGridY.at( vIdx )  = grY / ym0;
      m_padGridX.at( vIdx )      = grX / xm0;
      m_padGridY.at( vIdx )      = grY / ym0;
    } else {
      // If two maps to be crossed
      if ( xm0 < xm1 ) {
        // xm0 is the vertical map, xm1 is the horizontal map
        m_logVertGridX.at( vIdx ) = grX / xm0;
        m_logVertGridY.at( vIdx ) = grY / ym0;

        m_logHorizGridX.at( vIdx ) = grX / xm1;
        m_logHorizGridY.at( vIdx ) = grY / ym1;

        m_padGridX.at( vIdx ) = grX / xm0;
        m_padGridY.at( vIdx ) = grY / ym1;
      } else {
        // xm1 is the vertical map, xm0 is the horizontal map
        m_logVertGridX.at( vIdx ) = grX / xm1;
        m_logVertGridY.at( vIdx ) = grY / ym1;

        m_logHorizGridX.at( vIdx ) = grX / xm0;
        m_logHorizGridY.at( vIdx ) = grY / ym0;

        m_padGridX.at( vIdx ) = grX / xm1;
        m_padGridY.at( vIdx ) = grY / ym0;
      }
    }
  } else if ( readoutType.size() == 2 ) {
    // There are 2 different readouts
    if ( !readoutType.at( 0 ) ) {
      // First readout is anode readout xm0
      // must be two maps for Andode and cathode seperately
      if ( grX / xm0 > SgrX / xm1 ) {
        axm0  = xm0;
        axm1  = xm1;
        aym0  = ym0;
        aym1  = ym1;
        agrX  = grX;
        agrY  = grY;
        aSgrX = SgrX;
        aSgrY = SgrY;
      } else {
        axm0  = xm1;
        axm1  = xm0;
        aym0  = ym1;
        aym1  = ym0;
        agrX  = SgrX;
        agrY  = SgrY;
        aSgrX = grX;
        aSgrY = grY;
      }
    } else {
      // map over was reversed, otherwise the same
      if ( grX / xm1 > xm1 / xm0 ) {
        axm0  = xm1;
        axm1  = xm0;
        aym0  = ym1;
        aym1  = ym0;
        agrX  = grX;
        agrY  = grY;
        aSgrX = SgrX;
        aSgrY = SgrY;
      } else {
        axm0  = xm0;
        axm1  = xm1;
        aym0  = ym0;
        aym1  = ym1;
        agrX  = SgrX;
        agrY  = SgrY;
        aSgrX = grX;
        aSgrY = grY;
      }
    }

    m_logVertGridX.at( vIdx ) = agrX / axm0;
    m_logVertGridY.at( vIdx ) = agrY / aym0;

    m_logHorizGridX.at( vIdx ) = aSgrX / axm1;
    m_logHorizGridY.at( vIdx ) = aSgrY / aym1;

    m_padGridX.at( vIdx ) = agrX / axm0;
    m_padGridY.at( vIdx ) = aSgrY / aym1;

  } else {
    throw logic_error(
        fmt::format( "MuonChamberLayouty : Wrong size readout vector:: {} != from 1 or 2!!!!", readoutType.size() ) );
  }
  // End of FE work
  return StatusCode::SUCCESS;
}

std::optional<DeMuonDetector::TilePosition>
MuonChamberLayout::getXYZChamberTile( const LHCb::Detector::Muon::TileID& tile, bool toGlob ) {
  unsigned int chamberNum = getChamberNumber( tile );
  unsigned int station    = tile.station();
  unsigned int region     = tile.region();
  return getXYZChamber( station, region, chamberNum, toGlob );
}

std::optional<DeMuonDetector::TilePosition> MuonChamberLayout::getXYZChamber( const int& station, const int& region,
                                                                              const int& chamberNum, bool toGlob ) {
  return getXYZ( station, region, chamberNum, -1, toGlob );
}

std::optional<DeMuonDetector::TilePosition> MuonChamberLayout::getXYZ( const int& station, const int& region,
                                                                       const int& chamberNum, const int& gapNum,
                                                                       bool toGlob ) {

  int            encode, re( 0 );
  DeMuonChamber* muChamber;
  encode            = 276 * station + chamberNum;
  int MaxRegions[4] = { 12, 24, 48, 192 };
  if ( region ) {
    re = region;
    while ( re >= 1 ) {
      encode += MaxRegions[re - 1];
      re--;
    }
  }
  muChamber = m_ChVec.at( encode );

  if ( -1 == gapNum ) {

    // I'm intrested in the Chamber, not the gaps
    IGeometryInfoPlus* cInfo      = muChamber->geometryPlus();
    Gaudi::Transform3D vTransForm = cInfo->toGlobalMatrix();

    Gaudi::XYZVector  vtrans;
    Gaudi::Rotation3D vrota;
    vTransForm.GetDecomposition( vrota, vtrans );

    // get ILVolume pointer
    const ILVolume* logVol = cInfo->lvolume();
    // Get the solid
    const ISolid* solid = logVol->solid();
    // check these really are boxes (they ought to be!)
    const SolidBox* box = dynamic_cast<const SolidBox*>( solid );
    if ( !box ) { throw logic_error( "Could not cast gas gap solid to box" ); }

    const ISolid* GVso = muChamber->getFirstGasGap()->lvolume()->solid();
    // check these really are boxes (they ought to be!)
    const SolidBox* GVbox = dynamic_cast<const SolidBox*>( GVso );
    if ( !GVbox ) { throw logic_error( "Could not cast gas gap solid to box" ); }

    // Take the deltax and deltay from
    // Gas volume dimension
    double Dx = GVbox->xHalfLength();
    double Dy = GVbox->yHalfLength();
    // While take the deltaz from the whole chamber dimensions
    double Dz = box->zHalfLength();

    Gaudi::XYZPoint cnt( 0, 0, 0 );
    Gaudi::XYZPoint crn( Dx, Dy, Dz );
    if ( toGlob ) {
      localToglobal( cInfo, cnt, crn, Dx, Dy, Dz );
      return DeMuonDetector::TilePosition{ vtrans.x(), vtrans.y(), vtrans.z(), Dx, Dy, Dz };
    } else {
      return DeMuonDetector::TilePosition{ 0, 0, 0, Dx, Dy, Dz };
    }
  } else {

    // Going down to the needed gap
    // active volume is defined by the gas gaps, want first and last
    IGeometryInfoPlus* cInfo    = muChamber->geometryPlus();
    IPVolume*          gap      = muChamber->getGasGap( gapNum );
    IPVolume*          gapLayer = muChamber->getGasGapLayer( gapNum );

    if ( !gap ) { throw logic_error( "Could not read gas gaps from TDS" ); }

    // Found the needed gap
    const ISolid*   solid = gap->lvolume()->solid();
    const SolidBox* box   = dynamic_cast<const SolidBox*>( solid );
    if ( !box ) { throw logic_error( "Could not cast gas gap solid to box" ); }

    // Take the deltax and deltay from
    // Gas volume dimension
    double Dx = box->xHalfLength();
    double Dy = box->yHalfLength();
    // While take the deltaz from the whole chamber dimensions
    double Dz = box->zHalfLength();

    if ( toGlob ) {
      Gaudi::XYZPoint crn( Dx, Dy, Dz );
      Gaudi::XYZPoint pointInCh     = gapLayer->toMother( origin );
      Gaudi::XYZPoint pointInGlobal = muChamber->geometryPlus()->toGlobal( pointInCh );
      Gaudi::XYZPoint crnInCh       = gapLayer->toMother( crn );
      localToglobal( cInfo, pointInCh, crnInCh, Dx, Dy, Dz );
      return DeMuonDetector::TilePosition{ pointInGlobal.x(), pointInGlobal.y(), pointInGlobal.z(), Dx, Dy, Dz };
    } else {
      return DeMuonDetector::TilePosition{ 0, 0, 0, Dx, Dy, Dz };
    }
  }
}

std::optional<DeMuonDetector::TilePosition> MuonChamberLayout::getXYZPad( const LHCb::Detector::Muon::TileID& tile ) {
  // to find the x,y,z of the pad one must first find the chamber
  // to find the chamber one must know the tile of the chamber
  unsigned int station = tile.station();
  unsigned int region  = tile.region();

  // locate the chamber
  LHCb::Detector::Muon::TileID chamTile = m_layout[region].contains( tile );
  unsigned int                 chamb    = getChamberNumber( chamTile );

  // number of pads or logical channels in a chamber is :
  // layout.xGrid() / chamberLayoutX , padLayoutY / chamberLayoutY
  int xRatio = tile.layout().xGrid() / m_cgX.at( region );
  int yRatio = tile.layout().yGrid() / m_cgY.at( region );

  // need to work out offset within chamber
  // initally ignore the "reflectional" nature of the quarters
  int xOffset = tile.nX() - ( xRatio * chamTile.nX() );
  int yOffset = tile.nY() - ( yRatio * chamTile.nY() );

  // do the reflections of the tileID structure into Cartesian coordinates
  if ( 0 == tile.quarter() ) {
  } else if ( 1 == tile.quarter() ) {
    yOffset = yRatio - ( 1 + yOffset );
  } else if ( 2 == tile.quarter() ) {
    xOffset = xRatio - ( 1 + xOffset );
    yOffset = yRatio - ( 1 + yOffset );
  } else if ( 3 == tile.quarter() ) {
    xOffset = xRatio - ( 1 + xOffset );
  }

  // gap num??
  int gapNum = -1;
  // Here only gets the right deltas and offsets to current PAD.
  auto const pos = getXYZ( station, region, chamb, gapNum, false );
  if ( !pos ) { throw logic_error( fmt::format( "Failed to get chamber for pad tile {}", tile ) ); }

  // now make the calculations of x,y and z (and dx, dy and dz)
  // in chamber reference frame.
  double x = ( pos->x() - pos->dX() ) +
             ( ( static_cast<double>( xOffset ) / static_cast<double>( xRatio ) ) * ( 2.0 * pos->dX() ) );
  double y = ( pos->y() - pos->dY() ) +
             ( ( static_cast<double>( yOffset ) / static_cast<double>( yRatio ) ) * ( 2.0 * pos->dY() ) );
  double z = pos->z();

  double Dx = pos->dX() / ( static_cast<double>( xRatio ) );
  double Dy = pos->dY() / ( static_cast<double>( yRatio ) );

  // correct to the center of the pad
  x = x + Dx;
  y = y + Dy;

  Gaudi::XYZPoint cnt( x, y, z );
  Gaudi::XYZPoint crn( x + Dx, y + Dy, z + pos->dZ() );

  auto myChs = createChambersFromTile( { chamTile } );

  // Correct XYZ and Deltas for local to Global transformation
  double deltax = Dx;
  double deltay = Dy;
  double deltaz = pos->dZ();
  if ( !myChs.empty() ) {

    DeMuonChamber*     muChamber = myChs.at( 0 );
    IGeometryInfoPlus* cInfo     = muChamber->geometryPlus();
    // Those are not used in deltas computation
    auto cnt_global = cInfo->toGlobal( cnt );
    x               = cnt_global.x();
    y               = cnt_global.y();
    z               = cnt_global.z();

    localToglobal( cInfo, cnt, crn, deltax, deltay, deltaz );
  }

  return DeMuonDetector::TilePosition{ x, y, z, deltax, deltay, deltaz };
}

std::optional<DeMuonDetector::TilePosition>
MuonChamberLayout::getXYZLogical( const LHCb::Detector::Muon::TileID& tile ) {

  // If we get here then the logical strip is potenitally bigger than a chamber
  // if not then we can subcontract to getXYZPad directly

  // to find the x,y,z of the pad one must first find the chamber
  // to find the chamber one must know the tile of the chamber

  unsigned int station = tile.station();
  unsigned int region  = tile.region();

  if ( tile.layout().xGrid() >= m_layout[region].xGrid() && tile.layout().yGrid() >= m_layout[region].yGrid() ) {
    // ok logical channels are within a single chamber, delegate to getXYZpad
    return getXYZPad( tile );
  }
  if ( tile.layout().xGrid() >= m_layout[region].xGrid() && tile.layout().yGrid() < m_layout[region].yGrid() ) {
    // This logical channel crosses chambers in y, break it down into
    // chamber hight pads then combine the chamber extents later
    // number of tiles to make:
    int nTile = m_layout[region].yGrid() / tile.layout().yGrid();

    // width of pad (in x) is the same, hight (yGrid) that of the chamber
    MuonLayout tempPadLayout( tile.layout().xGrid(), m_layout[region].yGrid() );

    std::vector<LHCb::Detector::Muon::TileID> tempTiles;

    int i;
    for ( i = 0; i < nTile; i++ ) {
      int                          yTile = ( tile.nY() * nTile ) + i;
      LHCb::Detector::Muon::TileID tTile( station, tempPadLayout, region, tile.quarter(), tile.nX(), yTile );
      tempTiles.push_back( tTile );
    }

    // loop over the tiles and get max/min x,y,z of the combinations
    double                                                    xmin = 0.0;
    double                                                    xmax = 0.0;
    double                                                    ymin = 0.0;
    double                                                    ymax = 0.0;
    double                                                    zmin = 0.0;
    double                                                    zmax = 0.0;
    std::vector<LHCb::Detector::Muon::TileID>::const_iterator iTile;
    std::optional<DeMuonDetector::TilePosition>               pos{};
    for ( iTile = tempTiles.begin(); iTile != tempTiles.end(); iTile++ ) {
      pos = getXYZPad( *iTile );
      if ( !pos ) {
        throw logic_error( fmt::format( "Passed a temporary channel to getXYZPad: failed tile ID=", *iTile ) );
      }
      if ( iTile == tempTiles.begin() ) {
        xmin = pos->x();
        xmax = pos->x();
        ymin = pos->y();
        ymax = pos->y();
        zmin = pos->z();
        zmax = pos->z();
      } else {
        if ( xmax < pos->x() ) { xmax = pos->x(); }
        if ( xmin > pos->x() ) { xmin = pos->x(); }
        if ( ymax < pos->y() ) { ymax = pos->y(); }
        if ( ymin > pos->y() ) { ymin = pos->y(); }
        if ( zmax < pos->z() ) { zmax = pos->z(); }
        if ( zmin > pos->z() ) { zmin = pos->z(); }
      }
    }

    return DeMuonDetector::TilePosition{ ( xmax + xmin ) / 2.0,
                                         ( ymax + ymin ) / 2.0,
                                         ( zmax + zmin ) / 2.0,
                                         ( ( xmax - xmin ) / 2.0 ) + pos->dX(),
                                         ( ( ymax - ymin ) / 2.0 ) + pos->dY(),
                                         ( ( zmax - zmin ) / 2.0 ) + pos->dZ() };
  }
  throw std::out_of_range( "You requested a logical channel wider than a chamber, failing" );
}

std::optional<DeMuonDetector::TilePosition>
MuonChamberLayout::getXYZTwelfth( const LHCb::Detector::Muon::TileID& tile ) {
  unsigned int station = tile.station();
  unsigned int region  = tile.region();
  unsigned int quarter = tile.quarter();

  unsigned int twelfth;
  if ( 0 == quarter ) {
    if ( 1 == tile.nX() && 0 == tile.nY() ) {
      twelfth = 0;
    } else if ( 1 == tile.nX() && 1 == tile.nY() ) {
      twelfth = 1;
    } else {
      twelfth = 2;
    }
  } else if ( 1 == quarter ) {
    if ( 0 == tile.nX() && 1 == tile.nY() ) {
      twelfth = 3;
    } else if ( 1 == tile.nX() && 1 == tile.nY() ) {
      twelfth = 4;
    } else {
      twelfth = 5;
    }
  } else if ( 2 == quarter ) {
    if ( 1 == tile.nX() && 0 == tile.nY() ) {
      twelfth = 6;
    } else if ( 1 == tile.nX() && 1 == tile.nY() ) {
      twelfth = 7;
    } else {
      twelfth = 8;
    }
  } else {
    if ( 0 == tile.nX() && 1 == tile.nY() ) {
      twelfth = 9;
    } else if ( 1 == tile.nX() && 1 == tile.nY() ) {
      twelfth = 10;
    } else {
      twelfth = 11;
    }
  }

  double xMax( 0.0 ), xMin( 0.0 ), yMax( 0.0 ), yMin( 0.0 ), zMax( 0.0 ), zMin( 0.0 );

  int                                         chamCorner; // only need 3 corners here
  std::optional<DeMuonDetector::TilePosition> pos{};
  for ( chamCorner = 0; chamCorner < 4; chamCorner++ ) {
    int nChamber = getTwelfthCorner( region, twelfth, chamCorner );
    // get the chamber
    pos = getXYZChamber( station, region, nChamber, true );
    if ( !pos ) { throw logic_error( "Could not get corner chamber" ); }

    if ( 0 == chamCorner ) {
      xMax = pos->x();
      xMin = pos->x();
      yMax = pos->y();
      yMin = pos->y();
      zMax = pos->z();
      zMin = pos->z();
    } else {
      if ( pos->x() > xMax ) { xMax = pos->x(); }
      if ( pos->x() < xMin ) { xMin = pos->x(); }
      if ( pos->y() > yMax ) { yMax = pos->y(); }
      if ( pos->y() < yMin ) { yMin = pos->y(); }
      if ( pos->z() > zMax ) { zMax = pos->z(); }
      if ( pos->z() < zMin ) { zMin = pos->z(); }
    }
  } // end loop over chamCorner

  // strore the extent of the twelfth
  return DeMuonDetector::TilePosition{ ( xMax + xMin ) / 2.,
                                       ( yMax + yMin ) / 2.,
                                       ( zMax + zMin ) / 2.,
                                       ( ( xMax - xMin ) / 2. ) + pos->dX(),
                                       ( ( yMax - yMin ) / 2. ) + pos->dY(),
                                       ( ( zMax - zMin ) / 2. ) + pos->dZ() };
}

int MuonChamberLayout::getChamberNumber( const LHCb::Detector::Muon::TileID& tile ) {
  // look-up chamber number from tile
  // use the fact that the chamber numbering is constitent between stations
  unsigned int chamberNumber = 0; // counts from 1 as per TDS

  // Look inside the grid
  int fx( 0 ), fy( 0 );
  // Correct for quarter information
  if ( tile.quarter() == 0 ) {
    fx = tile.nX() + m_cgX.at( tile.region() ) * 2;
    fy = tile.nY() + m_cgY.at( tile.region() ) * 2;
  } else if ( tile.quarter() == 1 ) {
    fx = tile.nX() + m_cgX.at( tile.region() ) * 2;
    fy = -1 - tile.nY() + m_cgY.at( tile.region() ) * 2;
  } else if ( tile.quarter() == 2 ) {
    fx = -1 - tile.nX() + m_cgX.at( tile.region() ) * 2;
    fy = -1 - tile.nY() + m_cgY.at( tile.region() ) * 2;
  } else if ( tile.quarter() == 3 ) {
    fx = -1 - tile.nX() + m_cgX.at( tile.region() ) * 2;
    fy = tile.nY() + m_cgY.at( tile.region() ) * 2;
  }

  int enc       = fx + 4 * m_cgX.at( tile.region() ) * fy + m_offSet.at( tile.region() );
  chamberNumber = m_chamberGrid.at( enc );

  if ( chamberNumber > 1380 ) { logic_error( "Chamber number our of range for tile" ); }
  return chamberNumber;
}

int MuonChamberLayout::getTwelfthCorner( const int& region, const int& twelfth, const int& chamberNum ) {
  int xPos = 0;
  int yPos = 0;
  getTwelfthCornerIndex( region, twelfth, chamberNum, xPos, yPos );

  // get the actual chamber number
  std::vector<int> chamberNumber;
  chamberXY( xPos, yPos, 0, 0, region, chamberNumber );
  int ActChamber = chamberNumber.at( 0 );

  return ActChamber;
}

void MuonChamberLayout::getTwelfthCornerIndex( const int& region, const int& twelfth, const int& chamberNum, int& xPos,
                                               int& yPos ) {
  // chambers defining the corners of the twelfths
  // here the indexing in region, quater, twelfth
  // numbering clockwise:
  //
  //  +-------+ +-------+           ChamberNum
  //  | 4   3 | | 2   1 |           counts (no particular reason)
  //  |   +---+ +---+   |           +------+
  //  | 5 |         | 0 |           | 2  3 |
  //  +---+         +---+           | 0  1 |
  //  +---+         +---+           +------+
  //  | 6 |         |11 |
  //  |   +---+ +---+   |
  //  | 7   8 | | 9  10 |
  //  +-------+ +-------+
  // map to whole station numbering scheme

  if ( 0 == twelfth ) {
    xPos = 3 * m_cgX.at( region );
    yPos = 2 * m_cgY.at( region );
  } else if ( 1 == twelfth ) {
    xPos = 3 * m_cgX.at( region );
    yPos = 3 * m_cgY.at( region );
  } else if ( 2 == twelfth ) {
    xPos = 2 * m_cgX.at( region );
    yPos = 3 * m_cgY.at( region );
  } else if ( 3 == twelfth ) {
    xPos = 1 * m_cgX.at( region );
    yPos = 3 * m_cgY.at( region );
  } else if ( 4 == twelfth ) {
    xPos = 0;
    yPos = 3 * m_cgY.at( region );
  } else if ( 5 == twelfth ) {
    xPos = 0;
    yPos = 2 * m_cgY.at( region );
  } else if ( 6 == twelfth ) {
    xPos = 0;
    yPos = 1 * m_cgY.at( region );
  } else if ( 7 == twelfth ) {
    xPos = 0;
    yPos = 0;
  } else if ( 8 == twelfth ) {
    xPos = 1 * m_cgX.at( region );
    yPos = 0;
  } else if ( 9 == twelfth ) {
    xPos = 2 * m_cgX.at( region );
    yPos = 0;
  } else if ( 10 == twelfth ) {
    xPos = 3 * m_cgX.at( region );
    yPos = 0;
  } else {
    xPos = 3 * m_cgX.at( region );
    yPos = 1 * m_cgY.at( region );
  }

  // correct to positions of 1,2,3
  if ( 1 == chamberNum ) {
    xPos = xPos + ( m_cgX.at( region ) - 1 );
  } else if ( 2 == chamberNum ) {
    yPos = yPos + ( m_cgY.at( region ) - 1 );
  } else if ( 3 == chamberNum ) {
    xPos = xPos + ( m_cgX.at( region ) - 1 );
    yPos = yPos + ( m_cgY.at( region ) - 1 );
  }
}

void MuonChamberLayout::localToglobal( const IGeometryInfoPlus* gInfo, const Gaudi::XYZPoint& cent,
                                       const Gaudi::XYZPoint& corn, double& dx, double& dy, double& dz ) {
  auto cent_global = gInfo->toGlobal( cent );
  auto corn_global = gInfo->toGlobal( corn );

  dx = fabs( cent_global.x() - corn_global.x() );
  dy = fabs( cent_global.y() - corn_global.y() );
  dz = fabs( cent_global.z() - corn_global.z() );
}

std::vector<DeMuonChamber*>
MuonChamberLayout::createChambersFromTile( std::vector<LHCb::Detector::Muon::TileID> mytiles ) {

  std::vector<DeMuonChamber*> myChambers;
  int                         station, region, chamb;
  int                         MaxRegions[4] = { 12, 24, 48, 192 };
  int                         encode( 0 ), re( 0 );

  for ( const auto& idTile : mytiles ) {
    const long int cTile = idTile;
    station              = idTile.station();
    region               = idTile.region();
    chamb                = getChamberNumber( idTile );
    // Added protection against non existing chamber
    if ( cTile > 0 ) {
      encode = 276 * station + chamb;
      if ( region ) {
        re = region;
        while ( re >= 1 ) {
          encode += MaxRegions[re - 1];
          --re;
        }
      }
      myChambers.push_back( m_ChVec.at( encode ) );
    }
  }

  return myChambers;
}

std::vector<int> MuonChamberLayout::Tile2ChamberNum( const LHCb::Detector::Muon::TileID& tile ) {
  // OK how to make this work.....
  // first locate the station and region from the tile
  unsigned int station = tile.station();
  unsigned int region  = tile.region();

  // now compare the layout parameter to possible "levels"
  // currently chamber, logical channel, pad
  if ( m_cgX.at( region ) == tile.layout().xGrid() && m_cgY.at( region ) == tile.layout().yGrid() ) {
    // chambers
    return { getChamberNumber( tile ) };
  } else if ( m_padGridX[station * 4 + region] == tile.layout().xGrid() &&
              m_padGridY[station * 4 + region] == tile.layout().yGrid() ) {

    // muon pads and logical channels with a 1:1 mapping to pads
    const auto chamTile = m_layout[region].contains( tile );
    return { getChamberNumber( chamTile ) };
    ///    StatusCode sc = getXYZPad(tile,x,deltax,y,deltay,z,deltaz);
  } else if ( m_logHorizGridX[station * 4 + region] == tile.layout().xGrid() &&
              m_logHorizGridY[station * 4 + region] == tile.layout().yGrid() ) {
    // horizontal logical channels
    return Logical2ChamberNum( tile );
  } else if ( m_logVertGridX[station * 4 + region] == tile.layout().xGrid() &&
              m_logVertGridY[station * 4 + region] == tile.layout().yGrid() ) {
    // vertical logical channels
    return Logical2ChamberNum( tile );
  } else if ( 1 == tile.layout().xGrid() && 1 == tile.layout().yGrid() ) {
    // tile laying out Twelfths
    return Twelfth2ChamberNum( tile );
  } else {
    throw logic_error( fmt::format( "Did not understand the LHCb::Detector::Muon::TileID encoding xGrid={} yGrid={}, "
                                    "please check the muon geometry version.",
                                    tile.layout().xGrid(), tile.layout().yGrid() ) );
  }
}

std::vector<int> MuonChamberLayout::Logical2ChamberNum( const LHCb::Detector::Muon::TileID& tile ) {

  // If we get here then the logical strip is potenitally bigger than a chamber
  // if not then we can subcontract to getXYZPad directly

  // to find the x,y,z of the pad one must first find the chamber
  // to find the chamber one must know the tile of the chamber

  std::vector<int> chamberVect;

  unsigned int station = tile.station();
  unsigned int region  = tile.region();

  if ( tile.layout().xGrid() >= m_layout[region].xGrid() && tile.layout().yGrid() >= m_layout[region].yGrid() ) {

    // ok logical channels are within a single chamber
    LHCb::Detector::Muon::TileID chamTile = m_layout[region].contains( tile );
    unsigned int                 chamb    = getChamberNumber( chamTile );
    chamberVect.resize( 1 );
    chamberVect.at( 0 ) = chamb;
  } else if ( tile.layout().xGrid() >= m_layout[region].xGrid() && tile.layout().yGrid() < m_layout[region].yGrid() ) {
    // This logical channel crosses chambers in y, break it down into
    // chamber hight pads then combine the chamber extents later
    // number of tiles to make:
    int nTile = m_layout[region].yGrid() / tile.layout().yGrid(); // number of chamber crossed

    // width of pad (in x) is the same, hight (yGrid) that of the chamber
    MuonLayout tempPadLayout( tile.layout().xGrid(), m_layout[region].yGrid() );

    std::vector<LHCb::Detector::Muon::TileID> tempTiles;

    int i;
    for ( i = 0; i < nTile; i++ ) {
      int                          yTile = ( tile.nY() * nTile ) + i;
      LHCb::Detector::Muon::TileID tTile( station, tempPadLayout, region, tile.quarter(), tile.nX(), yTile );
      tempTiles.push_back( tTile );
    }

    // loop over the tiles and get the their chamber numbers
    for ( auto tile : tempTiles ) {
      const auto chamTile = m_layout[region].contains( tile );
      const auto chamb    = getChamberNumber( chamTile );
      chamberVect.push_back( chamb );
    }
  } else {
    throw std::out_of_range( "You requested a logical channel wider than a chamber, failing" );
  }
  return chamberVect;
}

std::vector<int> MuonChamberLayout::Twelfth2ChamberNum( const LHCb::Detector::Muon::TileID& tile ) {
  unsigned int     region  = tile.region();
  unsigned int     quarter = tile.quarter();
  std::vector<int> chamberVect;

  unsigned int twelfth;
  if ( 0 == quarter ) {
    if ( 1 == tile.nX() && 0 == tile.nY() ) {
      twelfth = 0;
    } else if ( 1 == tile.nX() && 1 == tile.nY() ) {
      twelfth = 1;
    } else {
      twelfth = 2;
    }
  } else if ( 1 == quarter ) {
    if ( 0 == tile.nX() && 1 == tile.nY() ) {
      twelfth = 3;
    } else if ( 1 == tile.nX() && 1 == tile.nY() ) {
      twelfth = 4;
    } else {
      twelfth = 5;
    }
  } else if ( 2 == quarter ) {
    if ( 1 == tile.nX() && 0 == tile.nY() ) {
      twelfth = 6;
    } else if ( 1 == tile.nX() && 1 == tile.nY() ) {
      twelfth = 7;
    } else {
      twelfth = 8;
    }
  } else {
    if ( 0 == tile.nX() && 1 == tile.nY() ) {
      twelfth = 9;
    } else if ( 1 == tile.nX() && 1 == tile.nY() ) {
      twelfth = 10;
    } else {
      twelfth = 11;
    }
  }

  int          chamCorner; // only need 3 corners here
  int          nChamber;
  unsigned int vdim = 0;
  for ( chamCorner = 0; chamCorner < 4; chamCorner++ ) {
    nChamber = getTwelfthCorner( region, twelfth, chamCorner );
    // get the chamber
    vdim++;
    chamberVect.resize( vdim );
    chamberVect.at( vdim ) = nChamber;
  }
  return chamberVect;
}
