/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#include "ANNSvcBase.h"
#include "Gaudi/Parsers/Factory.h"
#include "GaudiKernel/Service.h"
#include "GaudiKernel/ToStream.h"
#include "GaudiKernel/VectorMap.h"
#include "Kernel/IIndexedANNSvc.h"
#include "Kernel/IPropertyConfigSvc.h"
#include "Kernel/TCK.h"

namespace {
  using additionalIDs_t = std::map<std::string, std::map<std::string, unsigned int>>;

  static const std::array<Gaudi::StringKey, 5> s_majors{ "Hlt1SelectionID", "Hlt2SelectionID", "SpruceSelectionID",
                                                         "InfoID", "PackedObjectLocations" };
} // namespace

namespace Gaudi::Parsers {

  // Parser grammar
  template <typename Iterator, typename Skipper>
  struct Grammar_<Iterator, additionalIDs_t, Skipper> {
    // In this case, the type is a mapping type, so it requires the MapGrammar.
    // For other grammars see GaudiKernel/GrammarsV2.h
    typedef MapGrammar<Iterator, ::additionalIDs_t, Skipper> Grammar;
  };

  // Additional parser for our property
  StatusCode parse( additionalIDs_t& result, const std::string& input ) { return parse_( result, input ); }
} // namespace Gaudi::Parsers

// We also need to be able to print an object of our type as a string that both
// Python and our parser can understand,
namespace std {
  // This is an example valid for any mapping type.
  ostream& operator<<( ostream& s, const additionalIDs_t& m ) {
    return GaudiUtils::details::ostream_joiner( s << '{', m, ", ",
                                                []( ostream& os, const auto& i ) {
                                                  Gaudi::Utils::toStream(
                                                      i.second, Gaudi::Utils::toStream( i.first, os ) << ": " );
                                                } )
           << '}';
  }
} // namespace std

class TCKANNSvc : public ANNSvcBase {

  nlohmann::json generate_json( unsigned int ) const override {
    throw GaudiException( "generate_json not supported -- please use different service implementation",
                          __PRETTY_FUNCTION__, StatusCode::FAILURE );
    return {};
  }
  lumi_schema_t fetch_lumi( unsigned int, unsigned int ) const override {
    throw GaudiException( "fetch_lumi not supported -- please use different service implementation",
                          __PRETTY_FUNCTION__, StatusCode::FAILURE );
    return {};
  }
  map_t fetch( unsigned int tck, const Gaudi::StringKey& major ) const override {
    TCK _tck( tck );
    _tck.normalize();
    auto const* entry = m_cache.with_lock( [&]( std::map<TCK, const PropertyConfig*>& cache ) -> PropertyConfig const* {
      auto i = cache.find( _tck );
      if ( i != end( cache ) ) return i->second;
      // grab properties of child from config database...
      const ConfigTreeNode* tree = m_propertyConfigSvc->resolveConfigTreeNode(
          ConfigTreeNodeAlias::alias_type{ std::string( "TCK/" ) + _tck.str() } );
      if ( !tree ) {
        // If we could not resolve the (non-zero) TCK we have a problem
        error() << "Requested TCK " << _tck << " could not resolved. Returning an empty map... " << endmsg;
        return nullptr;
      }
      PropertyConfig::digest_type child = m_propertyConfigSvc->findInTree( tree->digest(), m_instanceName );
      if ( child.invalid() ) {
        error() << "Error finding configuration of " << m_instanceName << " for TCK " << _tck
                << " Returning an empty map... " << endmsg;
        return nullptr;
      }
      const PropertyConfig* config = m_propertyConfigSvc->resolvePropertyConfig( child );
      if ( !config ) {
        error() << "Error reading configuration of " << m_instanceName << " for TCK " << _tck
                << " Returning an empty map... " << endmsg;
        return nullptr;
      }
      auto status = cache.insert( { _tck, config } );
      if ( !status.second ) {
        error() << "Error updating cache for TCK " << _tck << " Returning an empty map... " << endmsg;
        return nullptr;
      }
      return status.first->second;
    } );

    if ( !entry ) return {};

    auto prop = std::find_if( std::begin( entry->properties() ), std::end( entry->properties() ),
                              [&]( const std::pair<std::string, std::string>& p ) { return major.str() == p.first; } );
    if ( prop == std::end( entry->properties() ) ) {
      error() << "Error finding requested major " << major << " in  configuration of " << m_instanceName << " for TCK "
              << _tck << " Returning an empty map... " << endmsg;
      return {};
    }
    // now use the property parser to do the hard work of converting the string rep into a map...
    std::map<std::string, int> map;
    auto                       status = Gaudi::Parsers::parse( map, prop->second );
    if ( status.isFailure() ) {
      error() << "Error interpreting requested major " << major << " in  configuration of " << m_instanceName
              << " for TCK " << _tck << " Returning an empty map... " << endmsg;
      return {};
    }
    GaudiUtils::VectorMap<unsigned int, std::string> result;
    auto                                             additionals = m_additionals.find( major.str() );
    bool                                             ok          = true;
    if ( additionals != end( m_additionals ) ) {
      for ( const auto& item : additionals->second ) { ok = ok && result.insert( item.second, item.first ).second; }
    }
    if ( !ok ) {
      error() << "Duplicate entries in additional map for  major " << major << " in  configuration of "
              << m_instanceName << " for TCK " << _tck << " Returning an empty map... " << endmsg;
      return {};
    }
    for ( const auto& item : map ) { ok = ok && result.insert( item.second, item.first ).second; }
    if ( !ok ) {
      error() << "Duplicate entries in map for  major " << major << " in  configuration of " << m_instanceName
              << " for TCK " << _tck << " Returning an empty map... " << endmsg;
      return {};
    }
    return result;
  }

  // properties
  Gaudi::Property<additionalIDs_t> m_additionals{
      this, "AdditionalIDs", {}, [this]( auto&& ) {
        if ( !m_additionals.empty() ) {
          warning() << "Additional IDs specified for " << m_instanceName << "." << endmsg
                    << "This is an expert option, and not intended for general use. Proceed at your own risk" << endmsg;
        };
      } };

  ServiceHandle<IPropertyConfigSvc> m_propertyConfigSvc{ this, "IPropertyConfigSvcInstance", "PropertyConfigSvc" };
  Gaudi::Property<std::string>      m_instanceName{ this, "InstanceName", "HltANNSvc" };
  mutable LHCb::cxx::SynchronizedValue<std::map<TCK, const PropertyConfig*>> m_cache; // TODO: flush cache if
                                                                                      // m_instanceName changes
public:
  using ANNSvcBase::ANNSvcBase;
};

DECLARE_COMPONENT( TCKANNSvc )
