###############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import json
import os
from pprint import pformat
from traceback import format_exc
from unittest import TestCase

FILENAME = f"{__name__}.root"
FILENAMEJSON = f"{__name__}.json"


def checkDiff(a, b):
    try:
        TestCase().assertEqual(a, b)
    except AssertionError as err:
        return str(err)


def config():
    from Configurables import ApplicationMgr

    from PyConf.Algorithms import EventCountAlg, Gaudi__Examples__IntDataProducer
    from PyConf.application import (
        ApplicationOptions,
        configure,
        configure_input,
        root_writer,
    )
    from PyConf.components import setup_component
    from PyConf.control_flow import CompositeNode

    options = ApplicationOptions(_enabled=False)
    # No data from the input is used, but something should be there for the configuration
    options.input_files = ["dummy_input_file_name.dst"]
    options.input_type = "ROOT"
    options.output_file = FILENAME
    options.output_type = "ROOT"
    options.data_type = "Upgrade"
    options.dddb_tag = "upgrade/dddb-20220705"
    options.conddb_tag = "upgrade/sim-20220705-vc-mu100"
    options.geometry_version = "run3/trunk"
    options.conditions_version = "master"
    options.simulation = True
    options.evt_max = 5
    options.monitoring_file = FILENAMEJSON
    # options.output_level = 2

    config = configure_input(options)

    app = ApplicationMgr()
    app.EvtSel = "NONE"  # ignore input configuration
    app.ExtSvc.append(
        config.add(
            setup_component(
                "LHCb__FSR__Sink",
                AcceptRegex=r"^EvtCounter1\.count$",
            )
        )
    )
    app.ExtSvc.append(
        config.add(
            setup_component(
                "LHCb__FSR__Sink",
                instance_name="FSRCount2",
                RecordName="SecondFSR",
                AcceptRegex=r"^EvtCounter2\.count$",
            )
        )
    )
    app.ExtSvc.append(
        config.add(
            setup_component(
                "LHCb__FSR__Sink",
                instance_name="FileSummaryRecord",
                AcceptRegex=r"^EvtCounter.\.count$",
            )
        )
    )

    producer = Gaudi__Examples__IntDataProducer()

    cf = CompositeNode(
        "test",
        [
            EventCountAlg(name="EvtCounter1"),
            EventCountAlg(
                name="EvtCounter2",
                OutputLevel=4,  # minor change to the configuration needed to please PyConf
            ),
            producer,
            root_writer(options.output_file, [producer.OutputLocation]),
        ],
    )
    app.OutStream.clear()
    config.update(configure(options, cf))

    # make sure the histogram file is not already there
    for name in [FILENAME, FILENAMEJSON]:
        if os.path.exists(name):
            os.remove(name)


def check(causes, result):
    result["root_output_file"] = FILENAME

    missing_files = [
        name for name in [FILENAME, FILENAMEJSON] if not os.path.exists(name)
    ]
    if missing_files:
        causes.append("missing output file(s)")
        result["missing_files"] = ", ".join(missing_files)
        return False

    expected_counters = {
        "EvtCounter1.count": {
            "empty": False,
            "nEntries": 5,
            "type": "counter:Counter:m",
        },
        "EvtCounter2.count": {
            "empty": False,
            "nEntries": 5,
            "type": "counter:Counter:m",
        },
    }
    expected = {
        "LHCb__FSR__Sink": ["EvtCounter1.count"],
        "SecondFSR": ["EvtCounter2.count"],
        "FileSummaryRecord": ["EvtCounter1.count", "EvtCounter2.count"],
    }

    try:
        import ROOT

        f = ROOT.TFile.Open(FILENAME)
        fsrs = {name: json.loads(str(getattr(f, name))) for name in expected}

    except Exception as err:
        causes.append("failure in root file")
        result["python_exception"] = result.Quote(format_exc())
        return False

    result["FSRs"] = result.Quote(pformat(fsrs))

    checking = "no check yet"
    try:
        tester = TestCase()
        for name, counters in expected.items():
            for counter in counters:
                checking = f"{name}[{counter}]"
                tester.assertEqual(expected_counters[counter], fsrs[name].get(counter))

    except AssertionError as err:
        causes.append("FSR content")
        result[f"FSR problem ({checking})"] = result.Quote(str(err))
        return False

    return True
