/*****************************************************************************\
* (c) Copyright 2000-2019 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#pragma once

// Include files
#include <ostream>

// Forward declarations

namespace LHCb {

  // Forward declarations

  /** @class VeloChannelID VeloChannelID.h
   *
   * This class describes a single channel of LHCb VELO Fully qualified
   * constructor: sets strip and sensor numbers and sensor type
   *
   * @author Chris Parkes, modified by David Hutchcroft
   *
   */

  class VeloChannelID final {
  public:
    /// Enumerates the 3 sensor types
    enum sensorType { Null = 0, RType = 1, PhiType = 2, PileUpType = 3 };

    /// constructor with sensor, strip, and sensor type
    constexpr VeloChannelID( unsigned int sensor, unsigned int strip, VeloChannelID::sensorType type ) {
      m_channelID = ( sensor << sensorBits ) | ( type << typeBits ) | strip;
    }

    /// constructor with sensor and strip number only (for backwards compatibility)
    constexpr VeloChannelID( unsigned int sensor, unsigned int strip )
        : VeloChannelID{ sensor, strip,
                         ( sensor < 64 )    ? VeloChannelID::sensorType::RType
                         : ( sensor < 128 ) ? VeloChannelID::sensorType::PhiType
                                            : VeloChannelID::sensorType::PileUpType } {}

    /// constructor with channelID
    constexpr explicit VeloChannelID( int id ) : m_channelID( id ) {}

    /// Default Constructor
    constexpr VeloChannelID() = default;

    /// Set sensor type bits
    void setType( const VeloChannelID::sensorType& type );

    /// Get sensor type
    [[nodiscard]] constexpr VeloChannelID::sensorType type() const {
      return sensorType( ( m_channelID & typeMask ) >> typeBits );
    }

    /// Is the channel in a pile up sensor
    [[nodiscard]] constexpr bool isPileUp() const { return VeloChannelID::sensorType::PileUpType == type(); }

    /// Is the channel in an R type sensor
    [[nodiscard]] constexpr bool isRType() const { return VeloChannelID::sensorType::RType == type(); }

    /// Is the channel in a phi type sensor
    [[nodiscard]] constexpr bool isPhiType() const { return VeloChannelID::sensorType::PhiType == type(); }

    /// cast
    [[nodiscard]] constexpr operator unsigned int() const { return m_channelID; }

    /// special serializer to ASCII stream
    std::ostream& fillStream( std::ostream& s ) const;

    /// Retrieve const  VELO Channel ID
    [[nodiscard]] constexpr unsigned int channelID() const { return m_channelID; }

    /// Update  VELO Channel ID
    VeloChannelID& setChannelID( unsigned int value ) {
      m_channelID = value;
      return *this;
    }

    /// Retrieve Strip number
    [[nodiscard]] constexpr unsigned int strip() const { return ( m_channelID & stripMask ) >> stripBits; }

    /// Update Strip number
    void setStrip( unsigned int value );

    /// Retrieve Sensor number
    [[nodiscard]] constexpr unsigned int sensor() const { return ( m_channelID & sensorMask ) >> sensorBits; }

    /// Update Sensor number
    void setSensor( unsigned int value );

    friend std::ostream& operator<<( std::ostream& str, const VeloChannelID& obj ) { return obj.fillStream( str ); }

  private:
    /// Offsets of bitfield channelID
    enum channelIDBits { stripBits = 0, sensorBits = 11, typeBits = 19 };

    /// Bitmasks for bitfield channelID
    enum channelIDMasks { stripMask = 0x7ffL, sensorMask = 0x7f800L, typeMask = 0x180000L };

    unsigned int m_channelID{ 0 }; ///< VELO Channel ID

  }; // class VeloChannelID

  inline std::ostream& operator<<( std::ostream& s, LHCb::VeloChannelID::sensorType e ) {
    switch ( e ) {
    case LHCb::VeloChannelID::Null:
      return s << "Null";
    case LHCb::VeloChannelID::RType:
      return s << "RType";
    case LHCb::VeloChannelID::PhiType:
      return s << "PhiType";
    case LHCb::VeloChannelID::PileUpType:
      return s << "PileUpType";
    default:
      return s << "ERROR wrong value " << int( e ) << " for enum LHCb::VeloChannelID::sensorType";
    }
  }

} // namespace LHCb

// -----------------------------------------------------------------------------
// end of class
// -----------------------------------------------------------------------------

inline void LHCb::VeloChannelID::setStrip( unsigned int value ) {
  auto val = (unsigned int)value;
  m_channelID &= ~stripMask;
  m_channelID |= ( ( ( (unsigned int)val ) << stripBits ) & stripMask );
}

inline void LHCb::VeloChannelID::setSensor( unsigned int value ) {
  auto val = (unsigned int)value;
  m_channelID &= ~sensorMask;
  m_channelID |= ( ( ( (unsigned int)val ) << sensorBits ) & sensorMask );
}

inline void LHCb::VeloChannelID::setType( const VeloChannelID::sensorType& type ) {
  m_channelID &= ~typeMask;
  m_channelID |= ( (unsigned long)( type ) << typeBits ) & typeMask;
}
