/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#ifndef TRACKKERNEL_TRACKDEFAULTPARTICLES_H
#define TRACKKERNEL_TRACKDEFAULTPARTICLES_H 1

#include "GaudiKernel/StatusCode.h"
#include <array>
#include <iomanip>
#include <stdexcept>
#include <stdint.h>

namespace LHCb {
  namespace Tr {
    class PID {
      enum class validated_pid_t { Electron = 0, Muon, Pion, Kaon, Proton, Xi, Omega, hypertriton };
      static constexpr std::array<double, 8> s_mass = { 0.51099891, 105.65837, 139.57018, 493.677,
                                                        938.27203,  1321.71,   1672.45,   2991.17 };

      validated_pid_t m_value;

      static constexpr validated_pid_t validate( int id ) {
        switch ( id ) {
        case 11:
          return validated_pid_t::Electron;
        case 13:
          return validated_pid_t::Muon;
        case 211:
          return validated_pid_t::Pion;
        case 321:
          return validated_pid_t::Kaon;
        case 2212:
          return validated_pid_t::Proton;
        case 3312:
          return validated_pid_t::Xi;
        case 3334:
          return validated_pid_t::Omega;
        case 1010010030:
          return validated_pid_t::hypertriton;
        default:
          throw std::runtime_error( "invalid PID" );
        }
      }

    public:
      constexpr explicit PID( const int id ) : m_value( validate( id ) ) {}

      constexpr explicit PID( validated_pid_t pid ) : m_value( pid ) {}

      PID() = default;
      // Make the creation of thes PID objects expressive in user code
      // by calling LHCb::Tr::PID::Electron()
      static constexpr PID Electron() { return PID{ validated_pid_t::Electron }; }

      static constexpr PID Muon() { return PID{ validated_pid_t::Muon }; }

      static constexpr PID Pion() { return PID{ validated_pid_t::Pion }; }

      static constexpr PID Kaon() { return PID{ validated_pid_t::Kaon }; }

      static constexpr PID Proton() { return PID{ validated_pid_t::Proton }; }

      static constexpr PID Xi() { return PID{ validated_pid_t::Xi }; }

      static constexpr PID Omega() { return PID{ validated_pid_t::Omega }; }

      static constexpr PID hypertriton() { return PID{ validated_pid_t::hypertriton }; }

      // Framwork functions allowing the use of PID inside a property
      friend const char* toString( PID pid ) {
        switch ( pid.m_value ) {
        case validated_pid_t::Electron:
          return "Electron";
        case validated_pid_t::Muon:
          return "Muon";
        case validated_pid_t::Pion:
          return "Pion";
        case validated_pid_t::Kaon:
          return "Kaon";
        case validated_pid_t::Proton:
          return "Proton";
        case validated_pid_t::Xi:
          return "Xi";
        case validated_pid_t::Omega:
          return "Omega";
        case validated_pid_t::hypertriton:
          return "hypertriton";
        default:
          throw std::runtime_error( "Calling toString on invalid PID" );
        }
      }

      friend std::ostream& toStream( const PID& pid, std::ostream& os ) {
        return os << std::quoted( toString( pid ), '\'' );
      }
      friend std::ostream& operator<<( std::ostream& os, const PID& pid ) { return toStream( pid, os ); }

      friend StatusCode parse( PID& pid, const std::string& in ) {
        for ( PID ref : { Electron(), Muon(), Pion(), Kaon(), Proton(), Xi(), Omega(), hypertriton() } ) {
          if ( in != toString( ref ) ) continue;
          pid = ref;
          return StatusCode::SUCCESS;
        }
        return StatusCode::FAILURE;
      }
      //----------------------------------------------------------

      constexpr double mass() const { return s_mass[static_cast<int>( m_value )]; }

      constexpr bool isElectron() const { return validated_pid_t::Electron == m_value; }

      constexpr bool isMuon() const { return validated_pid_t::Muon == m_value; }

      constexpr bool isPion() const { return validated_pid_t::Pion == m_value; }

      constexpr bool isKaon() const { return validated_pid_t::Kaon == m_value; }

      constexpr bool isProton() const { return validated_pid_t::Proton == m_value; }

      constexpr bool isXi() const { return validated_pid_t::Xi == m_value; }

      constexpr bool isOmega() const { return validated_pid_t::Omega == m_value; }

      constexpr bool ishypertriton() const { return validated_pid_t::hypertriton == m_value; }
    };
  } // namespace Tr
} // namespace LHCb

#endif
