/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#pragma once

#include "GaudiKernel/DataObject.h"
#include "GaudiKernel/IInterface.h"
#include "GaudiKernel/SmartRef.h"
#include "Relations/IRelation.h"
#include "Relations/Relation.h"
#include "Relations/RelationBase.h"
#include "Relations/RelationUtils.h"
#include <algorithm>
#include <utility>

namespace LHCb {
  /** @class Relation1D Relation1D.h Relations/Relation1D.h
   *
   *  @brief Implementation of ordinary unidirectional relations
   *
   *  Implementation of ordinary unidirectional relations
   *  from objects of type "FROM" to objects of type "TO".
   *
   *  Data types for "FROM" and "TO":
   *
   *   - either inherits form ContainedObject/DataObject
   *   - or have the templated specializations through
   *     ObjectTypeTraits structure with defined streamer operator
   *     to/from StreamBuffer class
   *
   *  @see DataObject
   *  @see ContainedObject
   *  @see ObjectTypeTraits
   *  @see StreamBuffer
   *  @warning for the current implementation the actual type of
   *           FROM should differ from the actual type of TO
   *
   *  @author Vanya Belyaev Ivan.Belyaev@itep.ru
   *  @date   25/01/2002
   */

  template <class FROM, class TO>
  class Relation1D : public DataObject, public Relations::BaseTable, public virtual IRelation<FROM, TO> {
  public:
    /// short cut for own     type
    using OwnType = Relation1D<FROM, TO>;
    /// short cut for inverse type
    using InvType = Relation1D<TO, FROM>;
    /// short cut for interface
    using IBase = IRelation<FROM, TO>;
    /// import "Range" type from the base
    using Range = typename IBase::Range;
    /// import "From"  type from the base
    using From = typename IBase::From;
    /// import "From"  type from the base
    using From_ = typename IBase::From_;
    /// import "To"    type from the base
    using To = typename IBase::To;
    /// import "To"    type from the base
    using To_ = typename IBase::To_;
    /// short cut for the actual implementation type
    using Base = typename Relations::Relation<FROM, TO>;
    // shortcut for "direct" interface
    using IDirect = typename IBase::DirectType;
    // shortcut for "inverse" interface
    using IInverse = typename IBase::InverseType;
    /// the actual type of the entry
    using Entry = typename IBase::Entry;

    /// Default constructor
    Relation1D() = default;
    /// constructor with reserve size
    explicit Relation1D( const size_t reserve ) : m_base( reserve ) {}
    /// constructor from any direct interface
    Relation1D( const IDirect& copy ) : m_base( copy ) {}
    /** constructor from "inverted interface"
     *  @param inv object to be inverted
     *  @param flag artificial argument to distinguisch from
     *  copy constructor
     */
    struct inverse_tag {};
    Relation1D( const IInverse& inv, inverse_tag ) : m_base( inv, typename Base::inverse_tag{} ) {}
    /// copy constructor
    Relation1D( const OwnType& c ) = default;
    /// move constructor
    Relation1D( OwnType&& m ) = default;

    /// assignement is disabled
    Relation1D& operator=( const Relation1D& ) = delete; // assignement is disabled
    Relation1D& operator=( Relation1D&& )      = delete; // assignement is disabled

    /// the type name
    [[nodiscard]] const std::string& type() const {
      static const std::string s_type( System::typeinfoName( typeid( OwnType ) ) );
      return s_type;
    }
    /// object identification (static method)
    static const CLID& classID() {
      static const CLID s_clid = Relations::clid( System::typeinfoName( typeid( OwnType ) ) );
      return s_clid;
    }
    /// object identification (virtual method)
    [[nodiscard]] const CLID& clID() const override { return classID(); }

    // major functional methods (fast, 100% inline)

    /// retrive all relations from the object (fast,100% inline)
    Range i_relations( From_ object ) const { return m_base.i_relations( object ); }
    /// retrive all relations from ALL objects (fast,100% inline)
    Range i_relations() const { return m_base.i_relations(); }
    /// make the relation between 2 objects (fast,100% inline method)
    StatusCode i_relate( From_ object1, To_ object2 ) { return i_add( { object1, object2 } ); }
    /// add entry
    StatusCode i_add( const Entry& entry ) { return m_base.i_add( entry ); }
    /// remove the concrete relation between objects (fast,100% inline method)
    StatusCode i_remove( From_ object1, To_ object2 ) { return m_base.i_remove( object1, object2 ); }
    /// remove all relations FROM the defined object (fast,100% inline method)
    StatusCode i_removeFrom( From_ object ) { return m_base.i_removeFrom( object ); }
    /// remove all relations TO the defined object (fast,100% inline method)
    StatusCode i_removeTo( To_ object ) { return m_base.i_removeTo( object ); }
    /// remove ALL relations form ALL  object to ALL objects  (fast,100% inline)
    StatusCode i_clear() { return m_base.i_clear(); }
    /// rebuild ALL relations form ALL  object to ALL objects(fast,100% inline)
    StatusCode i_rebuild() { return m_base.i_rebuild(); }
    /** make the relation between 2 objects (fast,100% inline method)
     *   - the call for i_sort() is mandatory
     */
    void i_push( From_ object1, To_ object2 ) { m_base.i_push( object1, object2 ); }
    /** (re)sort the table
     *   mandatory to use after i_push
     */
    void i_sort() { m_base.i_sort(); }

    // merge

    /** merge with the sorted range of relations
     *  @attention the range is assumed to be sorted!
     *  @param range the range to be added
     *  @return self-reference
     */
    Relation1D& merge( const Range& range ) {
      m_base.merge( range );
      return *this;
    }
    /** merge with the sorted range of relations
     *  @attention the range is assumed to be sorted!
     *  @param range the range to be added
     *  @return self-reference
     */
    Relation1D& imerge( const typename IInverse::Range& range ) {
      m_base.imerge( range );
      return *this;
    }
    /** merge with the sorted range of relations
     *  @attention the range is assumed to be sorted!
     *  @param range the range to be added
     *  @return self-reference
     */
    Relation1D& operator+=( const Range& range ) { return merge( range ); }

    // abstract methods from interface

    /// retrive all relations from the given object object
    Range relations( From_ object ) const override { return i_relations( object ); }
    /// retrive ALL relations from ALL objects
    Range relations() const override { return i_relations(); }
    /// make the relation between 2 objects
    StatusCode relate( From_ object1, To_ object2 ) override { return i_relate( object1, object2 ); }
    /// add the entry
    StatusCode add( const Entry& entry ) override { return i_add( entry ); }
    /// remove the concrete relation between objects
    StatusCode remove( From_ object1, To_ object2 ) override { return i_remove( object1, object2 ); }
    /// remove all relations FROM the defined object
    StatusCode removeFrom( From_ object ) override { return i_removeFrom( object ); }
    /// remove all relations TO the defined object
    StatusCode removeTo( To_ object ) override { return i_removeTo( object ); }
    /// remove ALL relations from ALL to ALL objects
    StatusCode clear() override { return i_clear(); }
    /// rebuild ALL relations from ALL  object to ALL objects
    StatusCode rebuild() override { return i_rebuild(); }
    /// update the object after POOL/ROOT reading
    StatusCode update() override { return i_rebuild(); }

    /// POOL identifier
    static std::string GUID() { return Relations::guid( classID() ); }

    /// Access the size of the relations
    std::size_t size() const { return m_base.size(); }

  private:
    /// the base table itself
    Base m_base; // the base table itself
  };

} // namespace LHCb
