###############################################################################
# (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
# Set of tests around XMLSummarySvc
# will read and write a Dst file
# Will fill some counters, and output the summary.xml

import os

from PyConf.Algorithms import LegacyEventCountAlg, RecordStream
from PyConf.application import (
    ApplicationOptions,
    configure,
    configure_input,
    create_or_reuse_rootIOAlg,
    root_writer,
)
from PyConf.control_flow import CompositeNode


def runTest(
    options, algs=[LegacyEventCountAlg(name="TestCounter")], insertWrongFile=False
):
    options.set_input_and_conds_from_testfiledb("TestXMLSummary")
    if insertWrongFile:
        options.input_files.insert(0, "YouCannotOpenMe.dst")
    options.output_file = os.environ.get("XMLSUMM_TEST_OUT", "RootDst.root")
    options.output_type = "ROOT"
    options.evt_max = 10
    options.xml_summary_svc = "CounterSummarySvc"
    options.root_ioalg_opts = {"StoreOldFSRs": True}
    config = configure_input(options)

    writer = root_writer(options.output_file, ["/Event/#1"])
    FSRWriter = RecordStream(
        name="FRWriter",
        Output=options.output_file,
        ItemList=["/FileRecords#999"],
        EvtDataSvc="FileRecordDataSvc",
        EvtConversionSvc="FileRecordPersistencySvc",
    )

    cf_node = CompositeNode(
        "Sequence",
        children=[create_or_reuse_rootIOAlg(options)] + algs + [writer, FSRWriter],
    )

    extra_config = configure(options, cf_node)

    # amend the XMLSummarySvc and set OutputLevel and UpdateFreq
    summarySvc = extra_config["XMLSummarySvc/CounterSummarySvc"]
    summarySvc.UpdateFreq = 1
    summarySvc.OutputLevel = 1
    config.update(extra_config)


def plain():
    """test plain configuration"""
    options = ApplicationOptions(_enabled=False)
    options.xml_summary_file = "summary.xml"
    runTest(options)


def exception():
    """test behavior when throwing an exception after 3 events"""
    from PyConf.Algorithms import ThrowingCounter

    options = ApplicationOptions(_enabled=False)
    options.xml_summary_file = "summary-exception.xml"
    runTest(options, [ThrowingCounter(name="Counter")])


def exit():
    """test behavior when exiting"""
    from PyConf.Algorithms import ExitingCounter

    options = ApplicationOptions(_enabled=False)
    options.xml_summary_file = "summary-exit.xml"
    runTest(options, [ExitingCounter(name="Counter")])


def readError():
    """test behavior in case of read error"""
    options = ApplicationOptions(_enabled=False)
    options.xml_summary_file = "summary-ReadError.xml"
    runTest(options, insertWrongFile=True)


def writeError():
    """test behavior in case of write error"""
    options = ApplicationOptions(_enabled=False)
    options.xml_summary_file = "summary-WriteError.xml"
    runTest(options)


def _log(causes, result, cause, error):
    causes.append(cause)
    result["xmlsummarykernel"] = result.Quote(error)


def _checkMemory(_, new, causes, result):
    stat = new.find("stat")
    if stat is None:
        _log(causes, result, f"XMLsummary content errorMissing child stat in usage")
    elif "useOf" not in stat.keys() or stat.get("useOf") != "MemoryMaximum":
        _log(
            causes,
            result,
            "XMLsummary content error",
            f"Expected useOf='MemoryMaximum' to be present In entry usage/stat",
        )


def _checkOutput(ref, new, causes, result):
    fref = ref.find("file")
    fnew = new.find("file")
    output_file = os.environ.get("XMLSUMM_TEST_OUT", "RootDst.root")
    if fnew is None:
        _log(
            causes, result, "XMLsummary content error", f"Missing child file in output"
        )
    elif output_file not in fnew.get("name"):
        _log(
            causes,
            result,
            "XMLsummary content error",
            f"Wrong output file {fnew.get('name')}, expected RootDst.root",
        )
    elif fnew.text != fref.text:
        _log(
            causes,
            result,
            "XMLsummary content error",
            f"Difference in output file usage : {fnew.text}, expected {fref.text}",
        )


def _checkCounters(ref, new, causes, result):
    def getTable(item):
        t = {}
        for child in item:
            t[child.get("name")] = child.text
        return t

    tref = getTable(ref)
    tnew = getTable(new)
    if tref != tnew:
        _log(
            causes,
            result,
            "XMLsummary different set of counters",
            f"Different counters : got {tnew}, expected {tref}",
        )


def _checkFiles(ref, new, causes, result):
    refFiles = {
        refChild.get("GUID"): refChild for refChild in ref if refChild.tag == "file"
    }
    newFiles = {
        newChild.get("GUID"): newChild for newChild in new if newChild.tag == "file"
    }
    for uuid in refFiles:
        if uuid not in newFiles:
            _log(
                causes,
                result,
                "missing input file",
                f"Input file with UUID {uuid} is missing",
            )
        else:
            _check(refFiles[uuid], newFiles[uuid], causes, result)
    for uuid in newFiles:
        if uuid not in refFiles:
            _log(
                causes,
                result,
                "extra input file",
                f"Got extra input file with UUID {uuid}",
            )


def _check(ref, new, causes, result):
    handlers = {
        "usage": _checkMemory,
        "output": _checkOutput,
        "input": _check,
        "counters": _checkCounters,
    }
    # check attributes are the same
    for attr in ref.keys():
        refValue = ref.get(attr)
        value = new.get(attr)
        if value is None:
            _log(
                causes,
                result,
                "XMLsummary missing attribute",
                f"Attribute {attr} is not present in tag {ref.tag}",
            )
        elif value != refValue:
            _log(
                causes,
                result,
                "XMLsummary wrong attribute",
                f"Attribute {attr} has unexpected value {value}, expected {refValue}",
            )
    # check content is the same
    if ref.text != new.text:
        _log(
            causes,
            result,
            "XMLsummary content error",
            f"In entry {ref.tag}, expected {ref.text}, got {new.text}",
        )
    # check children are ok
    hasFile = False
    for refChild in ref:
        tag = refChild.tag
        newChild = new.find(tag)
        if newChild is None:
            _log(
                causes,
                result,
                "XMLsummary missing entry",
                f"Entry {tag} is not present",
            )
        elif tag == "file":
            hasFile = True
        elif tag in handlers:
            handlers[tag](refChild, newChild, causes, result)
        else:
            if refChild.text != newChild.text:
                _log(
                    causes,
                    result,
                    "XMLsummary content error",
                    f"In entry {tag}, expected {refChild.text}, got {newChild.text}",
                )
    if hasFile:
        _checkFiles(ref, new, causes, result)


def check(refFile, outputFile, causes, result):
    import xml.etree.ElementTree as ET

    ref = ET.parse(refFile).getroot()
    try:
        new = ET.parse(outputFile).getroot()
    except FileNotFoundError:
        _log(causes, result, "XMLsummary missing", "XMLsummary missing")
    _check(ref, new, causes, result)
