/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

#include "Event/UTTELL1EventInfo.h"
#include <map>
#include <numeric>

//-----------------------------------------------------------------------------
// Implementation file for class : UTTELL1EventInfo
//
// 2008-07-17 : J Luisier
//-----------------------------------------------------------------------------

unsigned int LHCb::UTTELL1EventInfo::findPCN( const unsigned int beetle ) const {

  // get the pcn for a given beetle
  switch ( beetle ) {
  case 0:
    return pcnBeetle0();
  case 1:
    return pcnBeetle1();
  case 2:
    return pcnBeetle2();
  case 3:
    return pcnBeetle3();
  case 4:
    return pcnBeetle4();
  case 5:
    return pcnBeetle5();
  default:
    return 200u; //
  }
}

unsigned int LHCb::UTTELL1EventInfo::findIHeader( const unsigned int beetle ) const { // get the pcn for a given beetle
  switch ( beetle ) {
  case 0:
    return iHeaderBeetle0();
  case 1:
    return iHeaderBeetle1();
  case 2:
    return iHeaderBeetle2();
  case 3:
    return iHeaderBeetle3();
  case 4:
    return iHeaderBeetle4();
  case 5:
    return iHeaderBeetle5();
  default:
    return 200u; //
  }
}

unsigned int LHCb::UTTELL1EventInfo::pcnVote() const {

  std::map<unsigned int, unsigned int> pcns;
  for ( unsigned int iLink = 0; iLink < nBeetle; ++iLink )
    if ( linkEnabled( iLink ) ) { pcns[findPCN( iLink )] += 1; }
  // find the best
  struct Result {
    unsigned int majorityVote = 200;
    unsigned int maxValue     = 0;
  };
  return std::accumulate( pcns.begin(), pcns.end(), Result{},
                          []( Result r, const auto& i ) {
                            return i.second > r.maxValue ? Result{ i.first, i.second } : r;
                          } )
      .majorityVote;
}

bool LHCb::UTTELL1EventInfo::pcnConsistency() const {
  unsigned int pcnv( pcnVote() );

  return pcnv == pcnBeetle0() && pcnv == pcnBeetle1() && pcnv == pcnBeetle2() && pcnv == pcnBeetle3() &&
         pcnv == pcnBeetle4() && pcnv == pcnBeetle5();
}

bool LHCb::UTTELL1EventInfo::linkEnabled( const unsigned int iLink ) const { return !( OptLnkDisable() >> iLink & 1 ); }

bool LHCb::UTTELL1EventInfo::enabled() const {
  const unsigned int testValue = OptLnkDisable();
  return testValue != ( ( 1 << OptLnkDisableBits ) - 1 );
}

bool LHCb::UTTELL1EventInfo::noEvent( const unsigned int iLink ) const { return ( OptLnkNoEvt() >> iLink & 1 ); }

bool LHCb::UTTELL1EventInfo::noClock( const unsigned int iLink ) const { return ( OptLnkNoClock() >> iLink & 1 ); }

bool LHCb::UTTELL1EventInfo::TLKLinkLoss( const unsigned int iLink ) const { return ( tlkLnkLoss() >> iLink & 1 ); }

bool LHCb::UTTELL1EventInfo::syncRAMFull( const unsigned int iLink ) const { return ( SyncRAMFull() >> iLink & 1 ); }

bool LHCb::UTTELL1EventInfo::syncEvtSizeError( const unsigned int iLink ) const {
  return ( SyncEvtSizeError() >> iLink & 1 );
}

// fillstream method
std::ostream& LHCb::UTTELL1EventInfo::fillStream( std::ostream& s ) const {

  s << "Print Event info " << ChipAddr() << std::endl;
  for ( unsigned int i = 0; i < nBeetle; ++i ) {
    s << "Beetle: " << i << " PCN " << findPCN( i ) << " IHeader " << findIHeader( i ) << std::endl;
  } // i

  return s;
}
